#cython: language_level=3
# Copyright 2005-2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# See www.openfst.org for extensive documentation on this weighted
# finite-state transducer library.


#TODO(kbg): When/if PR https://github.com/cython/cython/pull/3358 is merged
# and we update third-party Cython up to or beyond a version that includes
# this, delete this file and instead use libcpp.utility.move.
cdef extern from * namespace "fst":
    """
    #include <type_traits>
    #include <utility>

    namespace fst {

    template <typename T>
    inline typename std::remove_reference<T>::type &&move(T &t) {
        return std::move(t);
    }

    template <typename T>
    inline typename std::remove_reference<T>::type &&move(T &&t) {
        return std::move(t);
    }

    }  // namespace fst
    """
    cdef T move[T](T)



'''
Created on 21 Jan 2010

@author: rfs
'''


import numpy as np
import collections
from itertools import combinations
import fstlib

""" Creates special transducers for special occasions ;) """
    
EDNAFULL = {('a','a'):5,
            ('t','t'):5,
            ('g','g'):5,
            ('c','c'):5,
            ('a','t'):-4,
            ('a','g'):-4,
            ('a','c'):-4,
            ('c','a'):-4,
            ('c','g'):-4,
            ('c','t'):-4,
            ('g','a'):-4,
            ('g','c'):-4,
            ('g','t'):-4,
            ('t','a'):-4,
            ('t','g'):-4,
            ('t','c'):-4}

EDITDNA = {('a','a'):0,
            ('t','t'):0,
            ('g','g'):0,
            ('c','c'):0,
            ('a','t'):1,
            ('a','g'):1,
            ('a','c'):1,
            ('c','a'):1,
            ('c','g'):1,
            ('c','t'):1,
            ('g','a'):1,
            ('g','c'):1,
            ('g','t'):1,
            ('t','a'):1,
            ('t','g'):1,
            ('t','c'):1}

#DNAEXP = {('A', 'A'): 2.8512358587881401, ('A', 'T'): 3.0396830090087774, ('T', 'T'): 1.8974875643790847, ('T', 'A'): 3.0396830090087774, ('C', 'A'): 3.7190999967568206, ('C', 'T'): 3.2662887680514583, ('G', 'G'): 2.8028092648826655, ('T', 'C'): 3.2662887680514583, ('G', 'A'): 3.6155593178159804, ('G', 'T'): 2.8398505365630147, ('A', 'G'): 3.6155593178159804, ('C', 'G'): 3.3569853295222702, ('C', 'C'): 2.4181654845880867, ('T', 'G'): 2.8398505365630147, ('G', 'C'): 3.3569853295222702, ('A', 'C'): 3.7190999967568206}
DNAEXP = {('A', 'A'): 2.3797654605525818, ('A', 'T'): 3.1532202470020141, ('T', 'T'): 2.3500179409534674, ('T', 'A'): 3.1532202470020141, ('C', 'A'): 3.1664024667929485, ('C', 'T'): 3.1572330934783825, ('G', 'G'): 2.392298397615908, ('T', 'C'): 3.1572330934783825, ('G', 'A'): 3.1828086852453756, ('G', 'T'): 3.1610199086227939, ('A', 'G'): 3.1828086852453756, ('C', 'G'): 3.1714878096208929, ('C', 'C'): 2.4101012294965733, ('T', 'G'): 3.1610199086227939, ('G', 'C'): 3.1714878096208929, ('A', 'C'): 3.1664024667929485}

def _adjust_scoring_matrix(scoring_matrix, invert = True):
    """the BioPython scoring matrices are not mirrored, i.e. if A->T is defined
    T->A isn't. This function fixes that. It additionally also multiplies all scores
    by -1 if 'invert = True' and turns all into lower case letters."""
    #unique_keys = set(map(lambda x:tuple(sorted(x)),scoring_matrix.keys()))
    new_scoring_matrix={}
        
    for key in scoring_matrix.keys():
        new_key = tuple([k.lower() for k in key])
        new_scoring_matrix[new_key] = scoring_matrix[key] * (-1 if invert else 1)
        new_scoring_matrix[new_key[::-1]] = new_scoring_matrix[new_key]
    
    return new_scoring_matrix
    
def create_three_state_global_alignment_fst(scoring_matrix, gap_open, gap_extend, invert_matrix = False, gap_symbol='-'):
    """creates a transducer for global alignment with
    specified gap open, extend and scoring costs. This FST 
    has 3 states and distinguishes between inserts and deletes """

    scoring_matrix = _adjust_scoring_matrix(scoring_matrix, invert_matrix)

    alphabet = set([i[0] for i in scoring_matrix.keys()])
    symbol_table = fstlib.SymbolTable()
    symbol_table.add_symbol(gap_symbol)
    for s in alphabet:
        symbol_table.add_symbol(s)
        
    myfst = fstlib.Fst(arc_type='standard')
    myfst.set_input_symbols(symbol_table)
    myfst.set_output_symbols(symbol_table)
    myfst.add_states(3)
    myfst.set_start(0)
    myfst.set_final(0,0)
    myfst.set_final(1,0)
    myfst.set_final(2,0)

    # starting state is match state
    # create match transitions to self
    myfst.add_arcs(0, [(s, t, scoring_matrix[(s,t)], 0) for s in alphabet for t in alphabet])

    # create gap open transitions to gap state [delete]
    myfst.add_arcs(0, [(s, gap_symbol, gap_open, 1) for s in alphabet])        
        
    # create gap open transitions to gap state [insert]
    myfst.add_arcs(0, [(gap_symbol, s, gap_open, 2) for s in alphabet])
        
    # create self transitions to gap states (gap extend) [delete]
    myfst.add_arcs(1, [(s, gap_symbol, gap_extend, 1) for s in alphabet])        
        
    # create self transitions to gap states (gap extend) [insert]
    myfst.add_arcs(2, [(gap_symbol, s, gap_extend, 2) for s in alphabet])

    # create match transitions back to match state
    myfst.add_arcs(2, [(s,t, scoring_matrix[(s,t)], 0) for s in alphabet for t in alphabet])
    myfst.add_arcs(1, [(s,t, scoring_matrix[(s,t)], 0) for s in alphabet for t in alphabet])
        
    # create delete to delete transitions
    myfst.add_arcs(1, [(gap_symbol, s, gap_open, 2) for s in alphabet])
    myfst.add_arcs(2, [(s, gap_symbol, gap_open, 1) for s in alphabet])
        
    return myfst, symbol_table

def create_kgram_fst(symbol_table, k, gap_cost=None):
    st = dict(symbol_table)
    myfst = fstlib.Fst(arc_type='log')
    myfst.set_input_symbols(symbol_table)
    myfst.set_output_symbols(symbol_table)
    myfst.add_states(k+1)
    myfst.set_start(0)
    myfst.set_final(k,0)
    gap = st[0]
    
    # state 0
    myfst.add_arcs(0, [(s, gap, 0, 0) for key,s in symbol_table if key!=0])
        
    for i in range(k):
        myfst.add_arcs(i, [(s,s, 0, i+1) for key,s in symbol_table if key!=0])
        if gap_cost is not None and i>0: ## gappy k-gram kernel
            myfst.add_arcs(i, [(s, gap, -np.log(gap_cost), i) for key,s in symbol_table if key!=0])

    # final state
    myfst.add_arcs(k, [(s, gap, 0, k) for key,s in symbol_table if key!=0])

    return myfst

def from_string(seq, final_weight=None, arc_weight=None, arc_type = 'standard', isymbols=None, osymbols=None):
    myfst = fstlib.Fst(arc_type=arc_type)
    startid = myfst.add_state()
    myfst.set_start(startid)
    myfst.set_input_symbols(isymbols)
    myfst.set_output_symbols(osymbols)
    for i,s in enumerate(seq):
        myfst.add_state()
        if arc_weight is None:
            arc_weight = fstlib.Weight.one(myfst.weight_type())
        else:
            if not isinstance(arc_weight, fstlib.Weight):
                arc_weight = fstlib.Weight(myfst.weight_type(), arc_weight)
        myfst.add_arc(i, (s, s, arc_weight, i+1))
    if final_weight is None:
        final_weight = fstlib.Weight.one(myfst.weight_type())
    else:
        if isinstance(final_weight, fstlib.Weight):
            pass
        else:
            final_weight = fstlib.Weight(myfst.weight_type(), final_weight)
    myfst.set_final(i+1, final_weight)
    return myfst

def from_array(seq, final_weight=None, arc_weight=None, arc_type = 'standard', symbols=None):
    myfst = fstlib.Fst(arc_type=arc_type)
    startid = myfst.add_state()
    myfst.set_start(startid)
    myfst.set_input_symbols(symbols)
    myfst.set_output_symbols(symbols)
    #for i,s in enumerate(seq):
    last_id = startid
    for i in range(seq.shape[1]): ## for each column
        current_id = myfst.add_state()        
        for j in range(seq.shape[0]): ## for each row
            entry = seq[j, i]
            if seq.ndim == 3:
                s, t = entry
            elif seq.ndim == 2:
                s = t = entry
            else:
                raise FSTFactoryError("Wrong number of dimensions in input array")
            if arc_weight is None:
                arc_weight = fstlib.Weight.one(myfst.weight_type())
            else:
                if not isinstance(arc_weight, fstlib.Weight):
                    arc_weight = fstlib.Weight(myfst.weight_type(), arc_weight)
            myfst.add_arc(last_id, (s, t, arc_weight, current_id))
        last_id = current_id

    if final_weight is None:
        final_weight = fstlib.Weight.one(myfst.weight_type())
    else:
        if isinstance(final_weight, fstlib.Weight):
            pass
        else:
            final_weight = fstlib.Weight(myfst.weight_type(), final_weight)
    myfst.set_final(last_id, final_weight)
    return myfst

def from_count_matrix(count_matrix, symbol_table, keys=None, arc_type = 'log', normalize=True):
    """ Creates a simple linear prob. FSA from a matrix of counts.
    The count matrix has a number of rows equal to the number of symbols and a number of
    columns equal to the length of the sequence. The keys parameter maps row indices in the count matrix
    to symbol table keys, i.e. keys[i] defines the symbol table key of the i-th row in the count matrix.
    If normalize is given, all column counts are normalized to sum to 1 (defines probability distribution). """
    st = dict(symbol_table)
    if keys is None:
        keys = list(st.keys())
    if count_matrix.shape[0] != len(keys):
        raise FSTFactoryError("Number of rows of count_matrix does not match symbol table or keys.")
    
    seq_length = count_matrix.shape[1]

    myfst = fstlib.Fst(arc_type)
    myfst.set_input_symbols(symbol_table)
    myfst.set_output_symbols(symbol_table)
    myfst.add_states(seq_length + 1)	
    myfst.set_start(0)
    myfst.set_final(seq_length, 0)

    count_sum = count_matrix.sum(0)
    if np.any(count_sum==0):
        raise FSTFactoryError("Column sumns must not be zero.")
                
    for i in range(0, seq_length):
        which = np.nonzero(count_matrix[:,i]>0)[0]
        for j in which:
            key = keys[j]
            prob = count_matrix[j,i]
            if normalize:
                prob = float(prob) / count_sum[i]
            if prob > 0:
                prob = - np.log(prob)
                myfst.add_arc(i, (key, key, prob, i+1))
        
    return myfst

def create_lattice(symbol_table, length, epsilons=False, arc_type='standard'):
    n = length+1
    fst = fstlib.Fst(arc_type=arc_type)
    fst.add_states(n)
    fst.set_start(0)
    fst.set_final(n-1, 0)
    fst.set_input_symbols(symbol_table)
    fst.set_output_symbols(symbol_table)
    for i in range(n-1):
        if epsilons:
            fst.add_arcs(i, [(s,s,0,i+1) for _,s in symbol_table])
        else:
            fst.add_arcs(i, [(s,s,0,i+1) for k,s in symbol_table if k != 0])

    return fst

def _hexdist(x, y):
    return abs(int(x, 16) - int(y, 16))

def _hexdist_signed(x, y):
    return int(x, 16) - int(y, 16)
    
def _hex_is_double(x, y):
    if x=="0" or y=="0": return(False)
    return (int(x, 16) / float(int(y, 16)) == 2)

def _hex_is_half(x, y):
    if x=="0" or y=="0": return(False)
    return (int(x, 16) / float(int(y, 16)) == 0.5)

def _hex_division(x, y):
    if x=="0" or y=="0": return(False)
    return (int(x, 16) / float(int(y, 16)))
    
class FSTFactoryError(Exception):
    def __init__(self, msg):
        self.message = msg

    def __str__(self):
        return self.message




import logging.config
import os

import yaml

with open(os.path.join(os.path.dirname(__file__), 'logging_conf.yaml'), 'rt') as f:
    config = yaml.safe_load(f.read())
logging.config.dictConfig(config)

IS_WINDOWS = os.name=="nt"

class Semiring:
    REAL = 'real'
    LOG = 'log'
    TROPICAL = 'standard'

DEF_GAP_SYMBOL = '-'
DEF_DELTA = 0.0009765625
MAX_INT32 = 2**31 - 1

from fstlib.cext.pywrapfst import *
from fstlib.core import *
from fstlib.ext import *
import fstlib.algos
import fstlib.factory
import fstlib.tools




import os
import pathlib
import subprocess
import time


def test_basic_import():
    "Testing that fstlib can be imported"
    import fstlib


def test_setup():
    "Testing that fstlib can be installed"
    process = subprocess.Popen(["python", "setup.py", "build_ext", "--inplace"],
                               stdout=subprocess.PIPE,
                               cwd=pathlib.Path(__file__).parent.parent.absolute())

    while process.poll() is None:
        # Process hasn't exited yet
        time.sleep(0.5)

    assert process.returncode == 0


def test_setup_deleted_so_cpp_files():
    "Testing that fstlib can be installed after deleting the existing .so and .cpp files"
    delete_process = subprocess.Popen(["rm", "fstlib/cext/*.so", "fstlib/cext/*.cpp"],
                                      stdout=subprocess.PIPE,
                                      cwd=pathlib.Path(__file__).parent.parent.absolute())

    b = pathlib.Path(__file__).parent.parent.absolute()
    a = os.listdir('fstlib/cext')
    c = delete_process.returncode == 0
    
    process = subprocess.Popen(["python", "setup.py", "build_ext", "--inplace"],
                               stdout=subprocess.PIPE,
                               cwd=pathlib.Path(__file__).parent.parent.absolute())

    while process.poll() is None:
        # Process hasn't exited yet
        time.sleep(0.5)

    c = process.returncode == 0

    assert process.returncode == 0



import copy
import logging
import os
from itertools import combinations

import Bio
import fstlib
import numpy as np
import pandas as pd

import medicc
from medicc import io, nj, tools, event_reconstruction

# prepare logger 
logger = logging.getLogger(__name__)


def main(input_df,
         asymm_fst,
         normal_name='diploid',
         input_tree=None,
         ancestral_reconstruction=True,
         chr_separator='X',
         prune_weight=0,
         allele_columns=['cn_a', 'cn_b'],
         wgd_x2=False,
         no_wgd=False,
         total_cn=False,
         n_cores=None,
         reconstruct_events=False):
    """ MEDICC Main Method """

    symbol_table = asymm_fst.input_symbols()

    ## Validate input
    logger.info("Validating input.")
    io.validate_input(input_df, symbol_table, normal_name=normal_name)

    ## Compile input data into FSAs stored in dictionaries
    logger.info("Compiling input sequences into FSAs.")
    FSA_dict = create_standard_fsa_dict_from_data(input_df, symbol_table, chr_separator)
    sample_labels = input_df.index.get_level_values('sample_id').unique()

    ## Reconstruct a tree
    if input_tree is None:
        ## Calculate pairwise distances
        logger.info("Calculating pairwise distance matrices")
        if n_cores is not None and n_cores > 1:
            pairwise_distances = parallelization_calc_pairwise_distance_matrix(sample_labels, 
                                                                        asymm_fst,
                                                                        FSA_dict,
                                                                        n_cores)
        else:
            pairwise_distances = calc_pairwise_distance_matrix(asymm_fst, FSA_dict)

        if (pairwise_distances == np.inf).any().any():
            affected_pairs = [(pairwise_distances.index[s1], pairwise_distances.index[s2])
                              for s1, s2 in zip(*np.where((pairwise_distances == np.inf)))]
            raise MEDICCError("Evolutionary distances could not be calculated for some sample "
                              "pairings. Please check the input data.\n\nThe affected pairs are: "
                              f"{affected_pairs}")

        logger.info("Inferring tree topology.")
        nj_tree = infer_tree_topology(
            pairwise_distances.values, pairwise_distances.index, normal_name=normal_name)
    else:
        logger.info("Tree provided, using it. No pairwise distance matrix is calculated!")

        pairwise_distances = pd.DataFrame(0, columns=FSA_dict.keys(), index=FSA_dict.keys())

        assert len([x for x in list(input_tree.find_clades()) if x.name is not None and 'internal' not in x.name]) == \
            len(np.unique(input_df.index.get_level_values('sample_id'))), \
            "Number of samples differs in input tree and input dataframe"
        assert np.all(
            np.sort([x.name for x in list(input_tree.find_clades()) if x.name is not None and 'internal' not in x.name]) ==
            np.sort(np.unique(input_df.index.get_level_values('sample_id')))), \
            "Input tree does not match input dataframe"
        
        # necessary for the way that reconstruct_ancestors is performed
        if ancestral_reconstruction:
            input_tree.root_with_outgroup([x for x in input_tree.root.clades if x.name != normal_name][0].name)

        nj_tree = input_tree


    final_tree = copy.deepcopy(nj_tree)

    if ancestral_reconstruction:
        logger.info("Reconstructing ancestors.")
        ancestors = medicc.reconstruct_ancestors(tree=final_tree,
                                                 samples_dict=FSA_dict,
                                                 fst=asymm_fst,
                                                 normal_name=normal_name,
                                                 prune_weight=prune_weight)

        ## Create and write output data frame with ancestors
        logger.info("Creating output copynumbers.")
        output_df = create_df_from_fsa(input_df, ancestors)

        ## Update branch lengths with ancestors
        logger.info("Updating branch lengths of final tree using ancestors.")
        update_branch_lengths(final_tree, asymm_fst, ancestors, normal_name)
    else:
        output_df = None

    nj_tree.root_with_outgroup(normal_name)
    final_tree.root_with_outgroup(normal_name)

    if ancestral_reconstruction and reconstruct_events:
        logger.info("Reconstructing events.")
        output_df, events_df = event_reconstruction.calculate_all_cn_events(
            final_tree, output_df, allele_columns, normal_name,
            wgd_x2=wgd_x2, no_wgd=no_wgd, total_cn=total_cn)
        if len(events_df) != final_tree.total_branch_length():
            faulty_nodes = []
            for node in final_tree.find_clades():
                if node.name is not None and node.name != normal_name and node.branch_length != 0 and node.branch_length != len(events_df.loc[node.name]):
                    faulty_nodes.append(node.name)
            logger.warn("Event recreation was faulty. Events in '_cn_events_df.tsv' will be "
                        f"incorrect for the following nodes: {faulty_nodes}")
    else:
        events_df = None

    return sample_labels, pairwise_distances, nj_tree, final_tree, output_df, events_df


def create_standard_fsa_dict_from_data(input_data,
                                       symbol_table: fstlib.SymbolTable,
                                       separator: str = "X") -> dict:
    """ Creates a dictionary of FSAs from input DataFrame or Series.
    The keys of the dictionary are the sample/taxon names. 
    If the input is a DataFrame, the FSA will be the concatenated copy number profiles of all allele columns"""

    fsa_dict = {}
    if isinstance(input_data, pd.DataFrame):
        logger.info('Creating FSA for pd.DataFrame with the following data columns:\n{}'.format(
            input_data.columns))
        def aggregate_copy_number_profile(cnp):
            return separator.join([separator.join(["".join(x.astype('str'))
                                                   for _, x in cnp[allele].groupby('chrom')]) for allele in cnp.columns])

    elif isinstance(input_data, pd.Series):
        logger.info('Creating FSA for pd.Series with the name {}'.format(input_data.name))
        def aggregate_copy_number_profile(cnp):
            return separator.join(["".join(x.astype('str')) for _, x in cnp.groupby('chrom')])

    else:
        raise MEDICCError("Input to function create_standard_fsa_dict_from_data has to be either"
                          "pd.DataFrame or pd.Series. \n input provided was {}".format(type(input_data)))
    
    for taxon, cnp in input_data.groupby('sample_id'):
        cn_str = aggregate_copy_number_profile(cnp)
        fsa_dict[taxon] = fstlib.factory.from_string(cn_str,
                                                     arc_type="standard",
                                                     isymbols=symbol_table,
                                                     osymbols=symbol_table)

    return fsa_dict


def create_phasing_fsa_dict_from_df(input_df: pd.DataFrame, symbol_table: fstlib.SymbolTable, separator: str = "X") -> dict:
    """ Creates a dictionary of FSAs from two allele columns (Pandas DataFrame).
    The keys of the dictionary are the sample/taxon names. """
    allele_columns = input_df.columns
    if len(allele_columns) != 2:
        raise MEDICCError("Need exactly two alleles for phasing.")

    fsa_dict = {}
    for taxon, cnp in input_df.groupby('sample_id'):
        allele_a = cnp[allele_columns[0]]
        allele_b = cnp[allele_columns[1]]
        cn_str_a = separator.join(["".join(x) for _,x in allele_a.groupby(level='chrom', sort=False)])
        cn_str_b = separator.join(["".join(x) for _,x in allele_b.groupby(level='chrom', sort=False)])
        encoded = np.array([list(zip(cn_str_a, cn_str_b)), list(zip(cn_str_b, cn_str_a))])
        fsa_dict[taxon] = fstlib.factory.from_array(encoded, symbols=symbol_table, arc_type='standard')
        fsa_dict[taxon] = fstlib.determinize(fsa_dict[taxon]).minimize()

    return fsa_dict

def phase(input_df: pd.DataFrame, model_fst: fstlib.Fst, reference_sample='diploid', separator: str = 'X') -> pd.DataFrame:
    """ Phases every FST against the reference sample. 
    Returns two standard FSA dicts, one for each allele. """
    
    diploid_fsa = medicc.tools.create_diploid_fsa(model_fst)
    phasing_dict = medicc.create_phasing_fsa_dict_from_df(input_df, model_fst.input_symbols(), separator)
    fsa_dict_a, fsa_dict_b, _ = phase_dict(phasing_dict, model_fst, diploid_fsa)
    output_df = medicc.create_df_from_phasing_fsa(input_df, [fsa_dict_a, fsa_dict_b], separator)

    # Phasing across chromosomes is random, so we need to swap haplotype assignment per chromosome
    # so that the higher ploidy haplotype is always cn_a
    output_df['width'] = output_df.eval('end+1-start')
    output_df['cn_a_width'] = output_df['cn_a'].astype(float) * output_df['width']
    output_df['cn_b_width'] = output_df['cn_b'].astype(float) * output_df['width']

    swap_haplotypes_ind = output_df.groupby(['sample_id', 'chrom'])[
        ['cn_a_width', 'cn_b_width']].mean().diff(axis=1).iloc[:, 1] > 0

    output_df = output_df.join(swap_haplotypes_ind.rename('swap_haplotypes_ind'), on=['sample_id', 'chrom'])
    output_df.loc[output_df['swap_haplotypes_ind'], ['cn_a', 'cn_b']] = output_df.loc[output_df['swap_haplotypes_ind'], ['cn_b', 'cn_a']].values
    output_df = output_df.drop(['width', 'cn_a_width', 'cn_b_width', 'swap_haplotypes_ind'], axis=1)

    return output_df

def phase_dict(phasing_dict, model_fst, reference_fst):
    """ Phases every FST against the reference sample. 
    Returns two standard FSA dicts, one for each allele. """
    fsa_dict_a = {}    
    fsa_dict_b = {}
    scores = {}
    left = (reference_fst * model_fst).project('output')
    right = (~model_fst * reference_fst).project('input')
    for sample_id, sample_fst in phasing_dict.items():
        phased_fst = fstlib.align(sample_fst, left, right).topsort()
        score = fstlib.shortestdistance(phased_fst, reverse=True)[phased_fst.start()]
        scores[sample_id] = float(score)
        fsa_dict_a[sample_id] = fstlib.arcmap(phased_fst.copy().project('input'), map_type='rmweight')
        fsa_dict_b[sample_id] = fstlib.arcmap(phased_fst.project('output'), map_type='rmweight')
    
    return fsa_dict_a, fsa_dict_b, scores


def create_df_from_fsa(input_df: pd.DataFrame, fsa, separator: str = 'X'):
    """ 
    Takes a single FSA dict or a list of FSA dicts and extracts the copy number profiles.
    The allele names are taken from the input_df columns and the returned data frame has the same 
    number of rows and row index as the input_df. """

    alleles = input_df.columns
    if not isinstance(fsa, dict):
        raise MEDICCError("fsa input to create_df_from_fsa has to be a dict"
                          "Input type is {}".format(type(fsa)))

    nr_alleles = len(alleles)
    samples = input_df.index.get_level_values('sample_id').unique()
    output_df = input_df.unstack('sample_id')

    # Create dict and concat later to prevent pandas PerformanceWarning
    internal_cns = dict()
    for node in fsa:
        if node in samples:
            continue
        cns = tools.fsa_to_string(fsa[node]).split(separator)
        if len(cns) % nr_alleles != 0:
            raise MEDICCError('For sample {} we have {} haplotype-specific chromosomes for {} alleles'
                              '\nnumber of chromosomes has to be divisible by nr of alleles'.format(node,
                                                                                                    len(cns),
                                                                                                    nr_alleles))
        nr_chroms = int(len(cns) // nr_alleles)
        for i, allele in enumerate(alleles):
            cn = list(''.join(cns[(i*nr_chroms):((i+1)*nr_chroms)]))
            internal_cns[(allele, node)] = cn

    internal_cns_df = pd.DataFrame(internal_cns, index=output_df.index)
    internal_cns_df.columns.names = ['allele', 'sample_id']
    output_df = (pd.concat([output_df, internal_cns_df], axis=1)
                 .stack('sample_id')
                 .reorder_levels(['sample_id', 'chrom', 'start', 'end'])
                 .sort_index())

    return output_df


def create_df_from_phasing_fsa(input_df: pd.DataFrame, fsas, separator: str = 'X'):
    """ 
    Takes a two FSAs dicts from phasing and extracts the copy number profiles.
    The allele names are taken from the input_df columns and the returned data frame has the same 
    number of rows and row index as the input_df. """

    alleles = input_df.columns
    if len(fsas) != 2:
        raise MEDICCError("fsas has to be of length 2")
    if not all([isinstance(fsa, dict) for fsa in fsas]):
        raise MEDICCError("all fsas entries have to be dicts")
    if fsas[0].keys() != fsas[1].keys():
        raise MEDICCError("fsas keys have to be the same")


    output_df = input_df.copy()[[]]
    output_df[alleles] = ''

    for sample in fsas[0].keys():
        cns_a = tools.fsa_to_string(fsas[0][sample]).split(separator)
        cns_b = tools.fsa_to_string(fsas[1][sample]).split(separator)
        if len(cns_a) != len(cns_b):
            raise MEDICCError(f"length of alleles is not the same for sample {sample}")

        output_df.loc[sample, alleles[0]] = list(''.join(cns_a))
        output_df.loc[sample, alleles[1]] = list(''.join(cns_b))

    # output_df = output_df.stack('sample_id')
    # output_df = output_df.reorder_levels(['sample_id', 'chrom', 'start', 'end']).sort_index()
    
    return output_df


def parallelization_calc_pairwise_distance_matrix(sample_labels, asymm_fst, FSA_dict, n_cores):
    try:
        from joblib import Parallel, delayed
    except ImportError:
        raise ImportError("joblib must be installed for parallelization")

    parallelization_groups = medicc.tools.create_parallelization_groups(len(sample_labels))
    parallelization_groups = [sample_labels[group] for group in parallelization_groups]
    logger.info("Running {} parallel runs on {} cores".format(len(parallelization_groups), n_cores))

    parallel_pairwise_distances = Parallel(n_jobs=n_cores)(delayed(calc_pairwise_distance_matrix)(
        asymm_fst, {key: val for key, val in FSA_dict.items() if key in cur_group}, True)
            for cur_group in parallelization_groups)

    pdm = medicc.tools.total_pdm_from_parallel_pdms(sample_labels, parallel_pairwise_distances)

    return pdm


def calc_pairwise_distance_matrix(model_fst, fsa_dict, parallel_run=True):
    '''Given a symmetric model FST and input FSAs in a form of a dictionary, output pairwise distance matrix'''

    samples = list(fsa_dict.keys())
    pdm = pd.DataFrame(0, index=samples, columns=samples, dtype=float)
    combs = list(combinations(samples, 2))
    ncombs = len(combs)

    for i, (sample_a, sample_b) in enumerate(combs):
        cur_dist = float(fstlib.kernel_score(model_fst, fsa_dict[sample_a], fsa_dict[sample_b]))
        pdm[sample_a][sample_b] = cur_dist
        pdm[sample_b][sample_a] = cur_dist

        if not parallel_run and (100*(i+1)/ncombs) % 10 == 0:  # log every 10%
            logger.info('%.2f%%', (i+1)/ncombs * 100)

    return pdm


def infer_tree_topology(pairwise_distances, labels, normal_name):
    if len(labels) > 2:
        tree = nj.NeighbourJoining(pairwise_distances, labels).tree

        tmpsearch = [c for c in tree.find_clades(name = normal_name)]
        normal_node = tmpsearch[0]
        root_path = tree.get_path(normal_node)[::-1]

        if len(root_path)>1:
            new_root = root_path[1]
            tree.root_with_outgroup(new_root)
    else:
        clade_ancestor = Bio.Phylo.PhyloXML.Clade(branch_length=0, name='internal_1')
        clade_ancestor.clades = [Bio.Phylo.PhyloXML.Clade(
            name=label, branch_length=0 if label == normal_name else 1) for label in labels]

        tree = Bio.Phylo.PhyloXML.Phylogeny(root=clade_ancestor)
        tree.root_with_outgroup(normal_name)

    return tree


def update_branch_lengths(tree, fst, ancestor_fsa, normal_name='diploid'):
    """ Updates the branch lengths in the tree using the internal nodes supplied in the FSA dict 
    """
    if len(ancestor_fsa) == 2:
        child_clade = [x for x in tree.find_clades() if x.name is not None and x.name != normal_name][0]
        child_clade.branch_length = float(fstlib.score(
            fst, ancestor_fsa[normal_name], ancestor_fsa[child_clade.name]))

    if not isinstance(ancestor_fsa, dict):
        raise MEDICCError("input ancestor_fsa to function update_branch_lengths has to be either a dict"
                          "provided type is {}".format(type(ancestor_fsa)))

    def _distance_to_child(fst, ancestor_fsa, sample_1, sample_2):
        return float(fstlib.score(fst, ancestor_fsa[sample_1], ancestor_fsa[sample_2]))

    for clade in tree.find_clades():
        if clade.name is None:
            continue
        children = clade.clades
        if len(children) != 0:
            for child in children:
                if child.name == normal_name:  # exception: evolution goes from diploid to internal node
                    logger.debug(f'Updating MRCA branch length from {child.name} to {clade.name}')
                    brs = _distance_to_child(fst, ancestor_fsa, child.name, clade.name)
                else:
                    logger.debug(f'Updating branch length from {clade.name} to {child.name}')
                    brs = _distance_to_child(fst, ancestor_fsa, clade.name, child.name)
                logger.debug(f'branch length: {brs}')
                child.branch_length = brs


def summarize_patient(tree, pdm, sample_labels, normal_name='diploid', events_df=None):
    """Calculate several summary values for the provided samples

    Args:
        tree (Bio.Phylo.Tree): Phylogenetic tree
        pdm (pandas.DataFrame): Pairwise distance matrix between the samples
        sample_labels (list): List of all samples
        normal_name (str, optional): Name of normal sample. Defaults to 'diploid'.
        events_df (pandas.DataFrame, optional): DataFrame containg all copy-number events. Defaults to None.

    Returns:
        pandas.DataFrame: Summary DataFrame
    """    
    branch_lengths = []
    for parent in tree.find_clades(terminal=False, order="level"):
        for child in parent.clades:
            if child.branch_length:
                branch_lengths.append(child.branch_length)
    
    nsamples = len(sample_labels)
    tree_length = np.sum(branch_lengths)
    avg_branch_length = np.mean(branch_lengths)
    min_branch_length = np.min(branch_lengths)
    max_branch_length = np.max(branch_lengths)
    median_branch_length = np.median(branch_lengths)
    # p_star = stats.star_topology_test(pdm)
    # p_clock = stats.molecular_clock_test(pdm,
    #                                      np.flatnonzero(np.array(sample_labels) == normal_name)[0])
    if events_df is None:
        wgd_status = "unknown"
    else:
        if "wgd" in events_df['type'].values:
            wgd_status = "WGD on branch " + \
                "and ".join(events_df.loc[events_df['type'] ==
                                          'wgd'].index.get_level_values('sample_id'))
        else:
            wgd_status = "no WGD"

    result = pd.Series({
        'nsamples': nsamples,
        'normal_name': normal_name,
        'tree_length': tree_length,
        'mean_branch_length': avg_branch_length,
        'median_branch_length': median_branch_length,
        'min_branch_length': min_branch_length,
        'max_branch_length': max_branch_length,
        # 'p_star': p_star,
        # 'p_clock': p_clock,
        'wgd_status': wgd_status,
    })
    
    return result


def detect_wgd(input_df, sample, total_cn=False, wgd_x2=False):
    wgd_fst = io.read_fst(total_copy_numbers=total_cn, wgd_x2=wgd_x2)
    no_wgd_fst = io.read_fst(no_wgd=True)

    diploid_fsa = medicc.tools.create_diploid_fsa(no_wgd_fst)
    symbol_table = no_wgd_fst.input_symbols()
    fsa_dict = medicc.create_standard_fsa_dict_from_data(input_df.loc[[sample]],
                                                         symbol_table, 'X')

    distance_wgd = float(fstlib.score(wgd_fst, diploid_fsa, fsa_dict[sample]))
    distance_no_wgd = float(fstlib.score(no_wgd_fst, diploid_fsa, fsa_dict[sample]))

    return distance_wgd != distance_no_wgd


class MEDICCError(Exception):
    pass



#cython: language_level=3
# Copyright 2005-2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# See www.openfst.org for extensive documentation on this weighted
# finite-state transducer library.


from libcpp.string cimport string

from cintegral_types cimport int8
from cintegral_types cimport int16
from cintegral_types cimport int32
from cintegral_types cimport int64
from cintegral_types cimport uint8
from cintegral_types cimport uint16
from cintegral_types cimport uint32
from cintegral_types cimport uint64


cdef extern from "<iostream>" namespace "std" nogil:

  cdef cppclass iostream:

    pass

  cdef cppclass istream(iostream):

    pass

  cdef cppclass ostream(iostream):

    pass


# We are ignoring openmodes for the moment.


cdef extern from "<fstream>" namespace "std" nogil:

  cdef cppclass ifstream(istream):

    ifstream(const string &)

  cdef cppclass ofstream(ostream):

    ofstream(const string &)


cdef extern from "<sstream>" namespace "std" nogil:

  cdef cppclass stringstream(istream, ostream):

    stringstream()

    string str()

    stringstream &operator<<(const string &)

    stringstream &operator<<(bool)

    # We define these in terms of the Google cintegral_types.

    stringstream &operator<<(int8)

    stringstream &operator<<(uint8)

    stringstream &operator<<(int16)

    stringstream &operator<<(uint16)

    stringstream &operator<<(int32)

    stringstream &operator<<(uint32)

    stringstream &operator<<(int64)

    stringstream &operator<<(uint64)

    stringstream &operator<<(double)

    stringstream &operator<<(long double)



import logging

import matplotlib as mpl
import matplotlib.colors as mcolors
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

logger = logging.getLogger(__name__)

COL_ALLELE_A = mpl.colors.to_rgba('orange')
COL_ALLELE_B = mpl.colors.to_rgba('teal')
COL_CLONAL = mpl.colors.to_rgba('lightgrey')
COL_NORMAL = mpl.colors.to_rgba('dimgray')
COL_GAIN = mpl.colors.to_rgba('red')
COL_WGD = mpl.colors.to_rgba('green')
COL_LOSS = mpl.colors.to_rgba('blue')
COL_CHR_LABEL = mpl.colors.to_rgba('grey')
COL_VLINES = '#1f77b4'
COL_MARKER_INTERNAL = COL_VLINES
COL_MARKER_TERMINAL = 'black'
COL_MARKER_NORMAL = 'green'
COL_SUMMARY_LABEL = 'grey'
COL_BACKGROUND = 'white'
COL_BACKGROUND_HATCH = 'lightgray'
COL_PATCH_BACKGROUND = 'white'
LINEWIDTH_COPY_NUMBERS = 2
LINEWIDTH_CHR_BOUNDARY = 1
LINEWIDTH_SEGMENT_BOUNDARY = 0.5
ALPHA_PATCHES = 0.15
ALPHA_PATCHES_WGD = 0.3
ALPHA_CLONAL = 0.3
BACKGROUND_HATCH_MARKER = '/////'
TREE_MARKER_SIZE = 40
YLABEL_FONT_SIZE = 8
YLABEL_TICK_SIZE = 6
XLABEL_FONT_SIZE = 10
XLABEL_TICK_SIZE = 8
CHR_LABEL_SIZE = 8
SMALL_SEGMENTS_LIMIT = 1e7


def plot_cn_profiles(
        input_df,
        input_tree=None,
        title=None,
        normal_name='diploid',
        mincn='auto',
        maxcn='auto',
        allele_columns=['cn_a', 'cn_b'],
        plot_summary=True,
        plot_subclonal_summary=True,
        plot_clonal_summary=False,
        hide_normal_chromosomes=False,
        ignore_segment_lengths=False,
        tree_width_scale=1,
        track_width_scale=1,
        height_scale=1,
        horizontal_margin_adjustment=0,
        close_gaps=False,
        show_small_segments=False,
        show_branch_support=False,
        hide_internal_nodes=False,
        chr_label_func=None,
        show_events_in_tree=False,
        show_events_in_cn=True,
        show_branch_lengths=True,
        detailed_xticks=False,
        clonal_transparant=False,
        label_func=None):
    
    df = input_df.copy()
    df[allele_columns] = df[allele_columns].astype(int)
    if len(allele_columns) > 2:
        logger.warn("More than two allels were provided ({})\n"
                    "Copy number tracks can only be plotted for 1 or 2 alleles".format(allele_columns))
    if len(np.setdiff1d(allele_columns, df.columns)):
        logger.warn("Some provided allele_columns are not in the dataframe"
                    "These are: {}".format(np.setdiff1d(allele_columns, df.columns)))

    if np.setdiff1d(['is_clonal', 'is_normal', 'is_gain', 'is_loss', 'is_wgd'], df.columns).size > 0:
        if input_tree is None:
            df[['is_normal', 'is_clonal', 'is_gain', 'is_loss', 'is_wgd']] = False
        else:
            df[['is_gain', 'is_loss', 'is_wgd']] = False
            cn_change = compute_cn_change(df=df[allele_columns], tree=input_tree, normal_name=normal_name)
            
            df.loc[(cn_change < 0).any(axis=1), 'is_loss'] = True
            df.loc[(cn_change > 0).any(axis=1), 'is_gain'] = True

            is_normal = ~df.unstack('sample_id')[['is_loss', 'is_gain', 'is_wgd']].any(axis=1)
            is_normal.name = 'is_normal'
            mrca = [x for x in input_tree.root.clades if x.name != normal_name][0].name
            is_clonal = ~df.loc[df.index.get_level_values('sample_id')!=mrca].unstack('sample_id')[['is_loss', 'is_gain', 'is_wgd']].any(axis=1)
            is_clonal.name = 'is_clonal'

            df = df.drop(['is_normal', 'is_clonal'], axis=1, errors='ignore')
            df = (df
                    .join(is_normal, how='inner')
                    .reorder_levels(['sample_id', 'chrom', 'start', 'end'])
                    .sort_index()
                    .join(is_clonal, how='inner')
                    .reset_index())
            df = (df
                    .set_index(['sample_id', 'chrom', 'start', 'end'])
                    .sort_index())




    if hide_normal_chromosomes:
        df = df.join(df.groupby('chrom')['is_normal'].all().to_frame('hide'))
        df = df.query("~hide").drop('hide', axis=1)

    if mincn=='auto':
        mincn = df[allele_columns].min().min()
    else:
        mincn = int(mincn)        
    if maxcn=='auto':
        maxcn = df[allele_columns].max().max()
    else:
        maxcn = int(maxcn)

    samples = df.index.get_level_values('sample_id').unique()
    if hide_internal_nodes:
        if input_tree is not None:
            samples = [x for x in df.index.get_level_values('sample_id').unique() if list(
                input_tree.find_clades(x))[0].is_terminal()]
        else:
            samples = [x for x in df.index.get_level_values('sample_id').unique() if 'internal_' not in x]
    nsamp = len(samples)
    nsegs = df.loc[samples[0], :].groupby('chrom').size()

    # == 2 because of diploid
    if input_tree is None or normal_name is None or nsamp == 2:
        plot_summary = False
        plot_subclonal_summary = False
        plot_clonal_summary = False

    if nsamp > 20:
        logger.warn('More than 20 samples were provided. Creating the copy number tracks will take '
                    'a long time to process and might crash. Best to use plot_cn_heatmap instead')

    df.reset_index(['start','end'], inplace=True)

    segment_lengths = df['end'] - df['start']
    df['small_segment'] = (segment_lengths) < SMALL_SEGMENTS_LIMIT
    if ignore_segment_lengths:
        segment_lengths.loc[:] = 1

    if close_gaps or ignore_segment_lengths:
        cur_df = df.loc[samples[0], :].reset_index()[['chrom', 'start', 'end']]

        cur_df['start_pos'] = np.cumsum(np.append([0], segment_lengths.loc[samples[0]]))[:-1]
        cur_df['end_pos'] = cur_df['start_pos'] + segment_lengths.loc[samples[0]].values

        cur_df.set_index(['chrom', 'start'], inplace=True)
        cur_df.drop('end', axis=1, inplace=True)

        df = df.join(cur_df, on=['chrom', 'start'])

    else:
        chrom_offset = df.loc[samples[0],:].reset_index().groupby('chrom', sort=False).max()['end']
        chrom_offset.dropna(inplace=True)
        chrom_offset.loc[:] = np.append(0, chrom_offset.cumsum().values[:-1])
        chrom_offset.name='offset'
        df = df.join(chrom_offset, on='chrom')
        df['start_pos'] = df['start'] + df['offset']
        df['end_pos'] = df['end'] + df['offset']

    df = df.reset_index().set_index(['sample_id', 'chrom', 'start', 'end'])

    if plot_summary or plot_subclonal_summary:
        agg_events = compute_cn_change(df=df[allele_columns], tree=input_tree, normal_name=normal_name)
        agg_events = agg_events.groupby(["chrom", "start", "end"], observed=True).sum()

        agg_events[['start_pos', 'end_pos', 'small_segment']
                       ] = df.loc[samples[0], ['start_pos', 'end_pos', 'small_segment']]

        mrca = [x for x in input_tree.root.clades if x.name != normal_name][0].name
        mrca_df = df.loc[mrca].copy()
        mrca_df.loc[:, allele_columns] = mrca_df.loc[:, allele_columns] - df.loc[normal_name, allele_columns]

    ## determine clade colors
    clade_colors = {}
    for sample in samples:
        ## determine if sample is terminal
        is_terminal = True
        if input_tree is not None:
            matches = list(input_tree.find_clades(sample))
            if len(matches)>0:
                clade = matches[0]
                is_terminal = clade.is_terminal()
        ## determine if sample is normal
        clade_colors[sample] = COL_MARKER_TERMINAL
        if not is_terminal:
            clade_colors[sample] = COL_MARKER_INTERNAL
        if sample == normal_name:
            clade_colors[sample] = COL_MARKER_NORMAL

    # define plot width and height and create figure
    track_width = nsegs.sum() * 0.2 * track_width_scale
    if input_tree is not None:
        tree_width = 2.5 * tree_width_scale ## in figsize
    else:
        tree_width = 0

    nrows = nsamp + int(plot_summary) + int(plot_subclonal_summary) + int(plot_clonal_summary)
    plotheight = 4 * 0.2 * nrows * height_scale
    plotwidth = tree_width + track_width
    tree_width_ratio = tree_width / plotwidth
    fig = plt.figure(figsize=(min(250, plotwidth), min(250, plotheight)), constrained_layout=True)
    if input_tree is None:
        gs = fig.add_gridspec(nrows, 1)
        cn_axes = [fig.add_subplot(gs[i]) for i in range(0, nrows)]
        y_posns = {sample: i for i, sample in enumerate(samples)} # as they appear
    else:
        gs = fig.add_gridspec(nrows, 2, width_ratios=[tree_width_ratio, 1-tree_width_ratio])
        tree_ax = fig.add_subplot(gs[0:nsamp, 0])
        cn_axes = [fig.add_subplot(gs[i]) for i in range(1,(2*(nrows))+1,2)]
        y_posns = _get_y_positions(
            input_tree, adjust=not hide_internal_nodes, normal_name=normal_name)
        y_posns = {clade.name: y_pos for clade, y_pos in y_posns.items(
            ) if clade.name is not None and clade.name != 'root'}
        plot_tree(input_tree, 
                  ax=tree_ax,
                  title=title,
                  normal_name=normal_name,
                  label_func=lambda x: '',
                  label_colors=clade_colors,
                  show_branch_support=show_branch_support,
                  show_events=show_events_in_tree,
                  show_branch_lengths=show_branch_lengths,
                  hide_internal_nodes=hide_internal_nodes)
    
    # Adjust the margin between the tree and cn tracks. Default is 0
    fig.set_constrained_layout_pads(w_pad=0, h_pad=0, hspace=0.0, wspace=horizontal_margin_adjustment)
    ## iterate over samples and plot the track
    for sample, group in df.groupby('sample_id'):
        if sample not in samples:
            continue
        index_to_plot = y_posns[sample] - 1
        plot_axis_labels = (index_to_plot == (nsamp-1))
        _plot_cn_profile(ax=cn_axes[index_to_plot],
                         label=label_func(sample) if label_func is not None else sample,
                         data=group,
                         mincn=mincn-1,
                         maxcn=maxcn+1,
                         alleles=allele_columns,
                         type='sample',
                         clonal_transparant=clonal_transparant,
                         chr_label_func=chr_label_func,
                         plot_xaxis_labels=plot_axis_labels if not (
                            plot_summary + plot_subclonal_summary + plot_clonal_summary) else False,
                         plot_yaxis_labels=True,
                         yaxis_label_color=clade_colors[sample],
                         show_small_segments=show_small_segments,
                         detailed_xticks=detailed_xticks,
                         show_events_in_cn=show_events_in_cn)

    if plot_clonal_summary:
        cur = mrca_df.copy()
        cur[['is_clonal', 'is_normal', 'is_gain', 'is_loss', 'is_wgd']] = False
        _plot_cn_profile(ax=cn_axes[nsamp],
                         label='clonal\nchanges',
                         data=cur,
                         mincn=mrca_df[allele_columns].min().min()-1,
                         maxcn=mrca_df[allele_columns].max().max()+1,
                         alleles=allele_columns,
                         detailed_xticks=detailed_xticks,
                         chr_label_func=chr_label_func,
                         type='summary',
                         show_small_segments=show_small_segments)
        cn_axes[nsamp].get_xaxis().set_visible(not (plot_summary or plot_subclonal_summary))

    if plot_summary:
        cur = agg_events.copy()
        cur[['is_clonal', 'is_normal', 'is_gain', 'is_loss', 'is_wgd']] = False
        _plot_cn_profile(ax=cn_axes[-1],
                         label='all\nchanges',
                         data=cur,
                         mincn=agg_events[allele_columns].min().min()-1,
                         maxcn=agg_events[allele_columns].max().max()+1,
                         alleles=allele_columns,
                         detailed_xticks=detailed_xticks,
                         chr_label_func=chr_label_func,
                         type='summary',
                         show_small_segments=show_small_segments)

    if plot_subclonal_summary:
        agg_events.loc[:, allele_columns] = agg_events[allele_columns] - mrca_df[allele_columns]
        cur = agg_events.copy()
        cur[['is_clonal', 'is_normal', 'is_gain', 'is_loss', 'is_wgd']] = False

        _plot_cn_profile(ax=cn_axes[-1 - int(plot_summary)],
                         label='subclonal\nchanges',
                         data=cur,
                         type='summary',
                         mincn=agg_events[allele_columns].min().min()-1,
                         maxcn=agg_events[allele_columns].max().max()+1,
                         alleles=allele_columns,
                         detailed_xticks=detailed_xticks,
                         chr_label_func=chr_label_func,
                         show_small_segments=show_small_segments)
        cn_axes[-1 - int(plot_summary)].get_xaxis().set_visible(not plot_summary)

    cn_axes[-1].set_xlabel('genome position', fontsize=XLABEL_FONT_SIZE)
    cn_axes[-1].xaxis.set_tick_params(labelsize=XLABEL_TICK_SIZE)
    cn_axes[-1].xaxis.offsetText.set_fontsize(XLABEL_TICK_SIZE)

    return fig


def _plot_cn_profile(ax, label, data, mincn, maxcn, alleles, type='sample',
                     plot_xaxis_labels=True, plot_yaxis_labels=True, chr_label_func=None,
                     clonal_transparant=True, yaxis_label_color='black', show_small_segments=False,
                     show_events_in_cn=True, detailed_xticks=False):
    ## collect line segments and background patches
    event_patches = []
    bkg_patches = []
    lines_a = []
    lines_b = []
    circles_a = []
    circles_b = []

    two_alleles = len(alleles) == 2

    alpha = []
    for idx, r in data.iterrows():
        lines_a.append([(r['start_pos'], r[alleles[0]]), 
                        (r['end_pos'], r[alleles[0]])])
        if two_alleles:
            lines_b.append([(r['start_pos'], r[alleles[1]]), (r['end_pos'], r[alleles[1]])])
        rect = mpl.patches.Rectangle((r['start_pos'], mincn), r['end_pos']-r['start_pos'],
                                     maxcn-mincn, edgecolor=None, facecolor=COL_PATCH_BACKGROUND, alpha=1)
        alpha.append(ALPHA_CLONAL if (clonal_transparant and r['is_clonal'] and type=='sample') else 1.0)
        bkg_patches.append(rect)

        if show_small_segments and r['small_segment']:
            circles_a.append((r['start_pos'] + 0.5*(r['end_pos'] - r['start_pos']), r[alleles[0]]))
            if two_alleles:
                circles_b.append(
                    (r['start_pos'] + 0.5*(r['end_pos'] - r['start_pos']), r[alleles[1]]))

        if show_events_in_cn and type=='sample':
            if r['is_normal']:
                rect = mpl.patches.Rectangle((r['start_pos'], mincn), r['end_pos']-r['start_pos'],
                                            maxcn-mincn, edgecolor=None, facecolor=COL_NORMAL, alpha=ALPHA_PATCHES)
                event_patches.append(rect)
            if r['is_wgd']:
                rect = mpl.patches.Rectangle((r['start_pos'], mincn), r['end_pos']-r['start_pos'],
                                             maxcn-mincn, edgecolor=None, facecolor=COL_WGD, alpha=ALPHA_PATCHES_WGD)
                event_patches.append(rect)
            if r['is_gain']:
                rect = mpl.patches.Rectangle((r['start_pos'], mincn), r['end_pos']-r['start_pos'],
                                            maxcn-mincn, edgecolor=None, facecolor=COL_GAIN, alpha=ALPHA_PATCHES)
                event_patches.append(rect)
            if r['is_loss']:
                rect = mpl.patches.Rectangle((r['start_pos'], mincn), r['end_pos']-r['start_pos'],
                                            maxcn-mincn, edgecolor=None, facecolor=COL_LOSS, alpha=ALPHA_PATCHES)
                event_patches.append(rect)

    events = mpl.collections.PatchCollection(event_patches, match_original=True, zorder=2)
    ax.add_collection(events)
    backgrounds = mpl.collections.PatchCollection(bkg_patches, match_original=True, zorder=1)
    ax.add_collection(backgrounds)
    plot_bkg = mpl.patches.Rectangle((0, 0), 1, 1, transform=ax.transAxes, facecolor=COL_BACKGROUND,
                                     edgecolor=COL_BACKGROUND_HATCH, zorder=0, hatch=BACKGROUND_HATCH_MARKER)
    ax.add_patch(plot_bkg)
    colors_a = np.array([COL_ALLELE_A] * len(lines_a))
    # clonal mutations
    colors_a[:, 3] = np.array(alpha)
    if two_alleles:
        colors_b = np.array([COL_ALLELE_B] * len(lines_b))
        colors_b[:, 3] = np.array(alpha)
        # a and b are overlapping
        colors_a[data[alleles[0]] == data[alleles[1]], 3] = 0.5
        colors_b[data[alleles[0]] == data[alleles[1]], 3] = 0.5
        colors = np.row_stack([colors_a, colors_b])
    else:
        colors = colors_a
    lc = mpl.collections.LineCollection(
        lines_a + lines_b, colors=colors, linewidth=LINEWIDTH_COPY_NUMBERS)
    ax.add_collection(lc)

    if type == 'summary':
        ax.axhline(0, color='grey', linestyle='--', zorder=1,
                   linewidth=LINEWIDTH_SEGMENT_BOUNDARY, alpha=0.75,)

    if len(circles_a) > 0:
        if two_alleles:
            a_b_overlap = (data.loc[data['small_segment']][alleles[0]] ==
                           data.loc[data['small_segment']][alleles[1]]).values
        else:
            a_b_overlap = circles_a * [False]
        ax.plot(np.array(circles_a)[~a_b_overlap, 0], np.array(circles_a)[~a_b_overlap, 1],
                'o', ms=3, color=COL_ALLELE_A, alpha=1., zorder=6)
        ax.plot(np.array(circles_a)[a_b_overlap, 0], np.array(circles_a)[a_b_overlap, 1],
                'o', ms=3, color=COL_ALLELE_A, alpha=0.5, zorder=6)
        if two_alleles:
            ax.plot(np.array(circles_b)[~a_b_overlap, 0], np.array(circles_b)[~a_b_overlap, 1],
                    'o', ms=3, color=COL_ALLELE_B, alpha=1., zorder=6)
            ax.plot(np.array(circles_b)[a_b_overlap, 0], np.array(circles_b)[a_b_overlap, 1],
                    'o', ms=3, color=COL_ALLELE_B, alpha=0.5, zorder=6)

    ax.autoscale()

    ## draw segment boundaries
    seg_bound_first = data['start_pos'].values[0]
    seg_bounds = data['end_pos'].values
    ax.vlines(np.append(seg_bound_first, seg_bounds), ymin=mincn, ymax=maxcn, ls='--', alpha=0.25,
              color=COL_VLINES, linewidth=LINEWIDTH_SEGMENT_BOUNDARY)

    ## draw chromosome boundaries
    chr_ends = data.reset_index().groupby('chrom').max()['end_pos']
    chr_ends.dropna(inplace=True)
    linex = chr_ends.values[:-1]  # don't plot last
    ax.vlines(linex, ymin=mincn, ymax=maxcn, color=COL_VLINES, linewidth=LINEWIDTH_CHR_BOUNDARY)

    ## draw chromosome labels
    chr_label_pos = chr_ends
    chr_label_pos.loc[:] = np.roll(chr_label_pos.values, 1)
    chr_label_pos.iloc[0] = seg_bound_first
    for chrom, pos in chr_label_pos.items():
        if chr_label_func is not None:
            chromtxt = chr_label_func(chrom)
        else:
            chromtxt = chrom

        ax.text(pos, maxcn-0.35, chromtxt, ha='left', va='top', color=COL_CHR_LABEL,
                fontweight='medium', fontsize=CHR_LABEL_SIZE)

    ## draw sample labels
    if plot_yaxis_labels:
        ax.set_ylabel(label, fontsize=YLABEL_FONT_SIZE, rotation=0, ha='right', va='center')
        ax.yaxis.label.set_color(yaxis_label_color)

    ## axis modifications
    if detailed_xticks:
        ax.set_xticks(np.arange(0, data['end_pos'].max(), 1e7))
    ax.get_xaxis().set_visible(plot_xaxis_labels)

    #ax.yaxis.set_major_locator(mpl.ticker.MaxNLocator(integer=True, prune='both', nbins=(maxcn-mincn)/2 + 1))
    ax.yaxis.set_major_locator(mpl.ticker.MaxNLocator(integer=True, prune='both', nbins=5))
    ax.yaxis.set_tick_params(labelsize=YLABEL_TICK_SIZE)
    ax.set_ylim(mincn, maxcn)
    ax.set_xlim(0, data['end_pos'].max())

def _get_x_positions(tree):
    """Create a mapping of each clade to its horizontal position.
    Dict of {clade: x-coord}
    """
    depths = tree.depths()
    # If there are no branch lengths, assume unit branch lengths
    if not max(depths.values()):
        depths = tree.depths(unit_branch_lengths=True)
    return depths

def _get_y_positions(tree, adjust=False, normal_name='diploid'):
    """Create a mapping of each clade to its vertical position.
    Dict of {clade: y-coord}.
Coordinates are negative, and integers for tips.
    """
    maxheight = tree.count_terminals()
    heights = {tip: maxheight -1 -i for i,
            tip in enumerate(reversed([x for x in tree.get_terminals() if x.name != normal_name]))}
    heights.update({list(tree.find_clades(normal_name))[0]: maxheight})

    # Internal nodes: place at midpoint of children
    def calc_row(clade):
        for subclade in clade:
            if subclade not in heights:
                calc_row(subclade)
        # Closure over heights
        heights[clade] = (heights[clade.clades[0]] + heights[clade.clades[-1]]) / 2.0

    if tree.root.clades:
        calc_row(tree.root)
        
    if adjust:
        pos = pd.DataFrame([(clade, val) for clade, val in heights.items()], columns=['clade','pos']).sort_values('pos')
        pos['newpos'] = 0
        count = 0
        for i in pos.index:
            if pos.loc[i,'clade'].name is not None and pos.loc[i,'clade'].name != 'root':
                count = count+1
            pos.loc[i, 'newpos'] = count

        pos.set_index('clade', inplace=True)
        heights = pos.to_dict()['newpos']

    return heights


def plot_tree(input_tree,
              label_func=None,
              title='',
              ax=None,
              output_name=None,
              normal_name='diploid',
              width_scale=1,
              height_scale=1,
              show_branch_lengths=True,
              show_branch_support=False,
              show_events=False,
              branch_labels=None,
              label_colors=None,
              hide_internal_nodes=False,
              marker_size=None,
              line_width=None,
              **kwargs):
    """Plot the given tree using matplotlib (or pylab).
    The graphic is a rooted tree, drawn with roughly the same algorithm as
    draw_ascii.
    Additional keyword arguments passed into this function are used as pyplot
    options. The input format should be in the form of:
    pyplot_option_name=(tuple), pyplot_option_name=(tuple, dict), or
    pyplot_option_name=(dict).
    Example using the pyplot options 'axhspan' and 'axvline'::
        from Bio import Phylo, AlignIO
        from Bio.Phylo.TreeConstruction import DistanceCalculator, DistanceTreeConstructor
        constructor = DistanceTreeConstructor()
        aln = AlignIO.read(open('TreeConstruction/msa.phy'), 'phylip')
        calculator = DistanceCalculator('identity')
        dm = calculator.get_distance(aln)
        tree = constructor.upgma(dm)
        Phylo.draw(tree, axhspan=((0.25, 7.75), {'facecolor':'0.5'}),
        ... axvline={'x':0, 'ymin':0, 'ymax':1})
    Visual aspects of the plot can also be modified using pyplot's own functions
    and objects (via pylab or matplotlib). In particular, the pyplot.rcParams
    object can be used to scale the font size (rcParams["font.size"]) and line
    width (rcParams["lines.linewidth"]).
    :Parameters:
        label_func : callable
            A function to extract a label from a node. By default this is str(),
            but you can use a different function to select another string
            associated with each node. If this function returns None for a node,
            no label will be shown for that node.
        do_show : bool
            Whether to show() the plot automatically.
        show_support : bool
            Whether to display confidence values, if present on the tree.
        ax : matplotlib/pylab axes
            If a valid matplotlib.axes.Axes instance, the phylogram is plotted
            in that Axes. By default (None), a new figure is created.
        branch_labels : dict or callable
            A mapping of each clade to the label that will be shown along the
            branch leading to it. By default this is the confidence value(s) of
            the clade, taken from the ``confidence`` attribute, and can be
            easily toggled off with this function's ``show_support`` option.
            But if you would like to alter the formatting of confidence values,
            or label the branches with something other than confidence, then use
            this option.
        label_colors : dict or callable
            A function or a dictionary specifying the color of the tip label.
            If the tip label can't be found in the dict or label_colors is
            None, the label will be shown in black.
    """

    try:
        import matplotlib.pyplot as plt
    except ImportError:
        try:
            import pylab as plt
        except ImportError:
            raise MEDICCPlotError(
                "Install matplotlib or pylab if you want to use draw."
            ) from None

    import matplotlib.collections as mpcollections
    if ax is None:
        nsamp = len(list(input_tree.find_clades()))
        plot_height = height_scale * nsamp * 0.25
        max_leaf_to_root_distances = np.max([np.sum([x.branch_length for x in input_tree.get_path(leaf)])
                            for leaf in input_tree.get_terminals()])
        plot_width = 5 + np.max([0, width_scale * np.log10(max_leaf_to_root_distances / 100) * 5])

        # maximum figure size is 250x250 inches
        fig, ax = plt.subplots(figsize=(min(250, plot_width), min(250, plot_height)))

    label_func=label_func if label_func is not None else lambda x: x

    # options for displaying label colors.
    if label_colors is not None:
        if callable(label_colors):
            def get_label_color(label):
                return label_colors(label)
        else:
            # label_colors is presumed to be a dict
            def get_label_color(label):
                return label_colors.get(label, "black")
    else:
        clade_colors = {}
        for sample in [x.name for x in list(input_tree.find_clades(''))]:
            ## determine if sample is terminal
            is_terminal = True
            matches = list(input_tree.find_clades(sample))
            if len(matches) > 0:
                clade = matches[0]
                is_terminal = clade.is_terminal()
            ## determine if sample is normal
            clade_colors[sample] = COL_MARKER_TERMINAL
            if not is_terminal:
                clade_colors[sample] = COL_MARKER_INTERNAL
            if sample == normal_name:
                clade_colors[sample] = COL_MARKER_NORMAL
        
        def get_label_color(label):
            return clade_colors.get(label, "black")

    if marker_size is None:
        marker_size = TREE_MARKER_SIZE
    marker_func=lambda x: (marker_size, get_label_color(x.name)) if x.name is not None else None

    ax.axes.get_yaxis().set_visible(False)
    ax.spines["right"].set_visible(False)
    ax.spines["left"].set_visible(False)
    ax.spines["top"].set_visible(False)
    ax.xaxis.set_major_locator(mpl.ticker.MaxNLocator(integer=True, prune=None))
    ax.xaxis.set_tick_params(labelsize=XLABEL_TICK_SIZE)
    ax.xaxis.label.set_size(XLABEL_FONT_SIZE)
    ax.set_title(title, x=0.01, y=1.0, ha='left', va='bottom',
                fontweight='bold', fontsize=16, zorder=10)
    x_posns = _get_x_positions(input_tree)
    y_posns = _get_y_positions(input_tree, adjust=not hide_internal_nodes, normal_name=normal_name)

    # Arrays that store lines for the plot of clades
    horizontal_linecollections = []
    vertical_linecollections = []

    # Options for displaying branch labels / confidence
    def value_to_str(value):
        if value is None or value == 0:
            return None
        elif int(value) == value:
            return str(int(value))
        else:
            return str(value)

    if not branch_labels:
        if show_branch_lengths:
            def format_branch_label(x): 
                return value_to_str(np.round(x.branch_length, 1)) if x.name != 'root' and x.name is not None else None
        else:
            def format_branch_label(clade):
                return None

    elif isinstance(branch_labels, dict):
        def format_branch_label(clade):
            return branch_labels.get(clade)
    else:
        if not callable(branch_labels):
            raise TypeError(
                "branch_labels must be either a dict or a callable (function)"
            )
        def format_branch_label(clade):
            return value_to_str(branch_labels(clade))

    if show_branch_support:
        def format_support_value(clade):
            if clade.name == 'root' or clade.name is None:
                return None
            try:
                confidences = clade.confidences
            # phyloXML supports multiple confidences
            except AttributeError:
                pass
            else:
                return "/".join(value_to_str(cnf.value) for cnf in confidences)
            if clade.confidence is not None:
                return value_to_str(clade.confidence)
            return None


    def draw_clade_lines(
        use_linecollection=False,
        orientation="horizontal",
        y_here=0,
        x_start=0,
        x_here=0,
        y_bot=0,
        y_top=0,
        color="black",
        lw=".1",
    ):
        """Create a line with or without a line collection object.
        Graphical formatting of the lines representing clades in the plot can be
        customized by altering this function.
        """
        if not use_linecollection and orientation == "horizontal":
            ax.hlines(y_here, x_start, x_here, color=color, lw=lw)
        elif use_linecollection and orientation == "horizontal":
            horizontal_linecollections.append(
                mpcollections.LineCollection(
                    [[(x_start, y_here), (x_here, y_here)]], color=color, lw=lw
                )
            )
        elif not use_linecollection and orientation == "vertical":
            ax.vlines(x_here, y_bot, y_top, color=color)
        elif use_linecollection and orientation == "vertical":
            vertical_linecollections.append(
                mpcollections.LineCollection(
                    [[(x_here, y_bot), (x_here, y_top)]], color=color, lw=lw
                )
            )

    def draw_clade(clade, x_start, color, lw):
        """Recursively draw a tree, down from the given clade."""
        x_here = x_posns[clade]
        y_here = y_posns[clade]
        # phyloXML-only graphics annotations
        if hasattr(clade, "color") and clade.color is not None:
            color = clade.color.to_hex()
        if hasattr(clade, "width") and clade.width is not None:
            lw = clade.width * plt.rcParams["lines.linewidth"]
        # Draw a horizontal line from start to here
        draw_clade_lines(
            use_linecollection=True,
            orientation="horizontal",
            y_here=y_here,
            x_start=x_start,
            x_here=x_here,
            color=color,
            lw=lw,
        )
        # Add node marker
        if marker_func is not None:
            marker = marker_func(clade)
            if marker is not None and clade is not None and not(hide_internal_nodes and not clade.is_terminal()):
                marker_size, marker_col = marker_func(clade)
                ax.scatter(x_here, y_here, s=marker_size, c=marker_col, zorder=3)

        # Add node/taxon labels
        label = label_func(str(clade.name))
        ax_scale = ax.get_xlim()[1] - ax.get_xlim()[0]

        if label not in (None, clade.__class__.__name__) and \
                not (hide_internal_nodes and not clade.is_terminal()):
            ax.text(
                x_here + min(0.02*ax_scale, 1),
                y_here,
                " %s" % label,
                verticalalignment="center",
                color=get_label_color(label),
            )
        # Add label above the branch
        conf_label = format_branch_label(clade)
        if conf_label:
            ax.text(
                0.5 * (x_start + x_here),
                y_here - 0.15,
                conf_label,
                fontsize="small",
                horizontalalignment="center",
            )
        # Add support below the branch
        if show_branch_support:
            support_value = format_support_value(clade)
            if support_value:
                ax.text(
                    0.5 * (x_start + x_here),
                    y_here + 0.25,
                    support_value + '%',
                    fontsize="small",
                    color='grey',
                    horizontalalignment="center",
                )
        # Add Events list
        if show_events and clade.events is not None:
            ax.text(
                0.5 * (x_start + x_here),
                y_here - 0.15,
                clade.events,
                fontsize="small",
                color=COL_MARKER_NORMAL,
                horizontalalignment="center",
            )
        if clade.clades:
            # Draw a vertical line connecting all children
            y_top = y_posns[clade.clades[0]]
            y_bot = y_posns[clade.clades[-1]]
            # Only apply widths to horizontal lines, like Archaeopteryx
            draw_clade_lines(
                use_linecollection=True,
                orientation="vertical",
                x_here=x_here,
                y_bot=y_bot,
                y_top=y_top,
                color=color,
                lw=lw,
            )
            # Draw descendents
            for child in clade:
                draw_clade(child, x_here, color, lw)

    if line_width is None:
        line_width = plt.rcParams["lines.linewidth"]
    draw_clade(input_tree.root, 0, "k", line_width)

    # If line collections were used to create clade lines, here they are added
    # to the pyplot plot.
    for i in horizontal_linecollections:
        ax.add_collection(i)
    for i in vertical_linecollections:
        ax.add_collection(i)

    ax.set_xlabel("branch length")
    ax.set_ylabel("taxa")

    # Add margins around the `tree` to prevent overlapping the ax
    xmax = max(x_posns.values())
    #ax.set_xlim(-0.05 * xmax, 1.25 * xmax)
    ax.set_xlim(-0.05 * xmax, 1.05 * xmax)
    # Also invert the y-axis (origin at the top)
    # Add a small vertical margin, but avoid including 0 and N+1 on the y axis
    ax.set_ylim(max(y_posns.values()) + 0.5, 0.5)

    # Parse and process key word arguments as pyplot options
    for key, value in kwargs.items():
        try:
            # Check that the pyplot option input is iterable, as required
            list(value)
        except TypeError:
            raise ValueError(
                'Keyword argument "%s=%s" is not in the format '
                "pyplot_option_name=(tuple), pyplot_option_name=(tuple, dict),"
                " or pyplot_option_name=(dict) " % (key, value)
            ) from None
        if isinstance(value, dict):
            getattr(plt, str(key))(**dict(value))
        elif not (isinstance(value[0], tuple)):
            getattr(plt, str(key))(*value)
        elif isinstance(value[0], tuple):
            getattr(plt, str(key))(*value[0], **dict(value[1]))

    if output_name is not None:
        plt.savefig(output_name + ".png", bbox_inches='tight')

    return plt.gcf()


def plot_cn_heatmap(input_df, final_tree=None, y_posns=None, cmax=8, total_copy_numbers=False,
                    alleles=['cn_a', 'cn_b'], tree_width_ratio=1, cbar_width_ratio=0.05, figsize=(20, 10),
                    tree_line_width=0.5, tree_marker_size=0, show_internal_nodes=False, title='',
                    tree_label_colors=None, tree_label_func=None, cmap='coolwarm', normal_name='diploid',
                    ignore_segment_lengths=False):

    input_df = input_df[alleles].copy()
    # if len(np.intersect1d(cur_sample_labels, input_df.index.get_level_values('sample_id').unique())) != len(cur_sample_labels):
    #     raise MEDICCPlotError("tree nodes and labels in dataframe are not the same")

    if not isinstance(alleles, list) and not isinstance(alleles, tuple):
        alleles = [alleles]
    nr_alleles = len(alleles)

    cmax = min(cmax, np.max(input_df[alleles].values.astype(int)))

    if final_tree is None:
        fig, axs = plt.subplots(figsize=figsize, ncols=1+nr_alleles, sharey=False,
                                gridspec_kw={'width_ratios': nr_alleles*[1] + [cbar_width_ratio]})

        cur_sample_labels = (input_df.index.get_level_values('sample_id').unique())
        if y_posns is None:
            y_posns = {s: i for i, s in enumerate(cur_sample_labels)}

        cn_axes = axs[:-1]
        cn_axes[0].set_title(title, x=0, y=1, ha='left', va='bottom', pad=20,
                             fontweight='bold', fontsize=16, zorder=10)
    else:
        fig, axs = plt.subplots(figsize=figsize, ncols=2+nr_alleles, sharey=False,
                                gridspec_kw={'width_ratios': [tree_width_ratio] + nr_alleles*[1] + [cbar_width_ratio]})
        tree_ax = axs[0]
        cn_axes = axs[1:-1]

        if show_internal_nodes:
            cur_sample_labels = np.array([x.name for x in list(final_tree.find_clades()) if x.name is not None])
        else:
            cur_sample_labels = np.array([x.name for x in final_tree.get_terminals()])

        y_posns = {k.name:v for k, v in _get_y_positions(final_tree, adjust=show_internal_nodes, normal_name=normal_name).items()}
        
        _ = plot_tree(final_tree, ax=tree_ax, normal_name=normal_name,
                      label_func=tree_label_func if tree_label_func is not None else lambda x: '',
                      hide_internal_nodes=(not show_internal_nodes), show_branch_lengths=False, show_events=False,
                      line_width=tree_line_width, marker_size=tree_marker_size,
                      title=title, label_colors=tree_label_colors)
        tree_ax.set_axis_off()
        tree_ax.set_axis_off()
        fig.set_constrained_layout_pads(w_pad=0, h_pad=0, hspace=0.0, wspace=100)
    cax = axs[-1]

    gaps = (input_df.loc[cur_sample_labels[0]].eval('start') - 
            np.roll(input_df.loc[cur_sample_labels[0]].eval('end'), 1)).values
    total_gaps = gaps[gaps>0].sum()
    if total_gaps > 1e8:
        logger.warn(f"Total of {total_gaps:.1e} bp gaps in the segmentation. These missing "
                    "segments are not reflected in the plot!")

    ind = [y_posns.get(x, -1) for x in cur_sample_labels]
    cur_sample_labels = cur_sample_labels[np.argsort(ind)]
    color_norm = mcolors.TwoSlopeNorm(vmin=0, vcenter=(2 if total_copy_numbers else 1), vmax=cmax)

    chr_ends = input_df.loc[cur_sample_labels[0]].copy()
    chr_ends['end_pos'] = np.cumsum([1]*len(chr_ends))
    chr_ends = chr_ends.reset_index().groupby('chrom').max()['end_pos']
    chr_ends = chr_ends.dropna()
    chr_ends = chr_ends.astype(int)


    if ignore_segment_lengths:
        x_pos = np.arange(len(input_df.loc[cur_sample_labels].astype(int).unstack('sample_id'))+1)
    else:
        x_pos = np.append([0], np.cumsum(input_df.loc[cur_sample_labels].astype(int).unstack(
            'sample_id').loc[:, (alleles[0])].loc[:, cur_sample_labels].eval('end+1-start').values))
    y_pos = np.arange(len(cur_sample_labels)+1)+0.5

    for ax, allele in zip(cn_axes, alleles):
        im = ax.pcolormesh(x_pos, y_pos,
                        input_df.loc[cur_sample_labels].astype(int).unstack(
                            'sample_id').loc[:, (allele)].loc[:, cur_sample_labels].values.T,
                        cmap=cmap,
                        norm=color_norm)

        for _, line in chr_ends.items():
            ax.axvline(x_pos[line], color='black', linewidth=0.75)
        
        xtick_pos = np.append([0], x_pos[chr_ends.values][:-1])
        xtick_pos = (xtick_pos + np.roll(xtick_pos, -1))/2
        xtick_pos[-1] += x_pos[-1]/2
        ax.set_xticks(xtick_pos)
        ax.set_xticklabels([x[3:] for x in chr_ends.index], ha='center', rotation=90, va='bottom')
        ax.tick_params(width=0)
        ax.xaxis.set_tick_params(labelbottom=False, labeltop=True, bottom=False)
        ax.set_yticks([])

    cax.pcolormesh([0, 1],
                   np.arange(0, cmax+2),
                   np.arange(0, cmax+1)[:, np.newaxis],
                   cmap=cmap,
                   norm=color_norm)

    cax.set_xticks([])
    cax.set_yticks(np.arange(0, cmax+1)+0.5)
    if np.max(input_df.values.astype(int)) > cmax:
        cax.set_yticklabels([str(x) + '+' if x == cmax else str(x)
                             for x in np.arange(0, cmax+1)], ha='left')
    else:
        cax.set_yticklabels(np.arange(0, cmax+1), ha='left')
    cax.yaxis.set_tick_params(left=False, labelleft=False, labelright=True)

    for ax in axs[:-1]:
        ax.set_ylim(len(cur_sample_labels)+0.5, 0.5)

    return fig


def compute_cn_change(df, tree, normal_name='diploid'):
    """Compute the copy-number changes per segment in all branches

    Args:
        df (pandas.DataFrame): DataFrame containing the copy-numbers of samples and internal nodes
        tree (Bio.Phylo.Tree): Phylogenetic tree
        normal_name (str, optional): Name of normal sample. Defaults to 'diploid'.

    Returns:
        pandas.DataFrame: DataFrame containing the copy-number changes
    """    
    cn_change = df.copy()
    alleles = cn_change.columns
    for allele in alleles:
        cn_change[allele] = cn_change[allele].astype('int')

    clades = [clade for clade in tree.find_clades(order = "postorder") if clade.name is not None and clade.name != normal_name]
    for clade in clades:
        for child in clade.clades:
            cn_change.loc[child.name, alleles] = cn_change.loc[child.name, alleles].values - cn_change.loc[clade.name, alleles].values
    cn_change.loc[clades[-1].name, alleles] = cn_change.loc[clades[-1].name, alleles].values - cn_change.loc[normal_name, alleles].values
    cn_change.loc[normal_name, alleles] = 0

    return cn_change


class MEDICCPlotError(Exception):
    pass



# distutils: language=c++
#cython: language_level=3
# See www.openfst.org for extensive documentation on this weighted
# finite-state transducer library.
#################################
# MODIFIED by the MEDICC2 authors
#################################


from libc.stdint cimport *


#cdef extern from "base/integral_types.h" nogil:
cdef extern from "<stdint.h>" nogil:

  ctypedef int8_t int8
  ctypedef int16_t int16
  ctypedef int32_t int32
  ctypedef int64_t int64
  ctypedef uint8_t uint8
  ctypedef uint16_t uint16
  ctypedef uint32_t uint32
  ctypedef uint64_t uint64



#cython: c_string_encoding=utf8, c_string_type=unicode, language_level=3, nonecheck=True
# distutils: language=c++

## Imports.

# Cython operator workarounds.
from cython.operator cimport address as addr       # &foo
from cython.operator cimport dereference as deref  # *foo
from cython.operator cimport preincrement as inc   # ++foo

## my imports
from libcpp.vector cimport vector
from .pywrapfst cimport *
from .cops cimport *
import numpy as np

def runme():
    print('This works')

cpdef MutableFst align(Fst model, Fst ifst1, Fst ifst2):
  
  cdef unique_ptr[fst.VectorFstClass] tfst
  tfst.reset(new fst.VectorFstClass(ifst1.arc_type()))

  cdef fst.WeightClass distance
  shortest_path(deref(model._fst), deref(ifst1._fst), deref(ifst2._fst), tfst.get())

  return _init_MutableFst(tfst.release())

cpdef Weight score_std(Fst model, Fst ifst1, Fst ifst2):
  distance = shortest_distance_std(deref(model._fst), deref(ifst1._fst), deref(ifst2._fst))
  retval = Weight(model._fst.get().WeightType(), distance.ToString())
  return retval


cpdef Weight score_log(Fst model, Fst ifst1, Fst ifst2):
  distance = shortest_distance_log(deref(model._fst), deref(ifst1._fst), deref(ifst2._fst))
  retval = Weight(model._fst.get().WeightType(), distance.ToString())
  return retval

cpdef Weight kernel_score_std(Fst model, Fst ifst1, Fst ifst2):
  distance = kernel_score_std_impl(deref(model._fst), deref(ifst1._fst), deref(ifst2._fst))
  retval = Weight(model._fst.get().WeightType(), distance.ToString())
  return retval





'''
Created on 17 May 2010

@author: rfs

FST algorithms
'''

import array
import functools
import logging

import numpy as np

import fstlib.core
#from fstlib.core import *
import fstlib.tools

log = logging.getLogger(__name__)

class FstAlgoError(Exception):
	pass

class ConfusionNetworkFromLattice(object):
	""" Expects lattice in the form of a FST object. 
	Requires lattice to be topologically sorted
	(arcs are always directed from lower to higher state ids). It might require (not sure)
   the initial state to have index 0. """
	def __init__(self, lattice):
		self.lattice = lattice
		self.state2set = None ## state set ids for each state in lattice
		self.statesets = None ## list of lists of state ids contained in the state sets
		self.edgesets = None ## new edges between state sets
		
	def run(self, collapse_arcs=True):
		log.info('Constructing confusion network...')
		lattice = self.lattice
		outfst = fstlib.core.Fst(arc_type = self.lattice.arc_type())
		outfst.set_input_symbols(self.lattice.input_symbols())
		outfst.set_output_symbols(self.lattice.output_symbols())
		timeslices = fstlib.shortestdistance(fstlib.arcmap(fstlib.arcmap(fstlib.arcmap(self.lattice, map_type='rmweight'), map_type='times', weight=1), map_type='to_std'))
		statemap = np.array([int(float(t)) for t in timeslices])
		newstates = np.array(outfst.add_states(np.unique(statemap).shape[0]))
		outfst.set_start(newstates[0])
		for s in newstates:
			weights = [self.lattice.final(t) for t in np.flatnonzero(statemap==s) if self.lattice.is_final(t)]
			if len(weights)>0: ## newstate is final if any of the merged states were final (have non-inf weights)
				total_weight = functools.reduce(fstlib.times, weights)
				outfst.set_final(s, total_weight)

		for state in self.lattice.states():
			newstate = statemap[state]
			for arc in self.lattice.arcs(state):
				newnextstate = statemap[arc.nextstate]
				outfst.add_arc(newstate, (arc.ilabel, arc.olabel, arc.weight, newnextstate))

		if collapse_arcs:
			outfst=fstlib.determinize(outfst)
		return outfst

class PosteriorDecoding(object):
	""" expects pFSA in binary (i.e. openfst) format """
	def __init__(self, ifst):
		self.ifst = ifst

	def run(self):
		alpha = fstlib.shortestdistance(self.ifst)
		beta = fstlib.shortestdistance(self.ifst, reverse=True)
		
		outfst = fstlib.core.Fst(arc_type=self.ifst.arc_type())
		outfst.set_input_symbols(self.ifst.input_symbols())
		outfst.set_output_symbols(self.ifst.output_symbols())
		
		for state in self.ifst.states():
			newstate = outfst.add_state()
			if self.ifst.is_final(state):
				outfst.set_final(newstate, self.ifst.final(state))
			if state != newstate:
				raise FSTError('State mapping error') 
				## this must not happen; I'm not sure it can currently due to the way the states are handled
		initial_state = self.ifst.start()
		outfst.set_start(initial_state)
		for state in self.ifst.states():
			for arc in self.ifst.arcs(state):
				posterior = float(alpha[state]) + float(arc.weight) + float(beta[arc.nextstate]) - float(beta[initial_state])
				outfst.add_arc(state, (arc.ilabel, arc.olabel, posterior, arc.nextstate))
		return outfst

class AbstractCountDP(object):
	def __init__(self, fst, trans_cache=None, state_depth_cache=None):
		""" constructor """
		self.fst = fst
		
		## if transition cache was given use it, otherwise create it
		if trans_cache == None:
			self._build_trans_cache()
		else:
			self._trans_cache = trans_cache
		
		## if state depth cache was given use it, otherwise create it
		if state_depth_cache == None:
			self._build_state_depth_cache()
		else:
			self._state_depth_cache = state_depth_cache
			
		assert(fst != None)

	def _build_trans_cache(self):
		#self._trans_cache = numpy.empty( (len(self.fst.states), len(self.fst.states)), dtype="object")
		self._trans_cache = [0] * len(self.fst.states)
		self._trans_cache = [[i] * len(self.fst.states) for i in self._trans_cache]
		
		for i in range(0, len(self.fst.states)):
			from_transs = [t for t in self.fst.arcs if t.state_from.id == i]
			for j in range(0, len(self.fst.states)):
				transs = len([t for t in from_transs if t.state_to.id == j])
				self._trans_cache[i][j] = transs
				
	def _build_state_depth_cache(self):
		""" tries to speed up the algo """
		self._state_depth_cache=list()
		depth = 0
		self._build_state_depth_cache_recurse(self.fst.states[0], depth)
		
	def _build_state_depth_cache_recurse(self, state, depth):
		""" recursion function of the pre inspection """
		
		if len(self._state_depth_cache) <= depth:
			self._state_depth_cache.append(set())
		self._state_depth_cache[depth].add(state.id)
		
		targets = list(set([t.state_to for t in state.get_outgoing_transitions()]))
		for s in targets:
			self._build_state_depth_cache_recurse(s, depth + 1)

	def max_seq_len(self):
		return len(self._state_depth_cache)-1

	def get_transition_cache(self):
		return self._trans_cache
	
	def get_state_depth_cache(self):
		return self._state_depth_cache
				
		
class ForwardCountDP(AbstractCountDP):
	""" Dynamic programming forward algorithm.
	This is not in log space and very generic. """
	
	def __init__(self, fst, trans_cache=None, state_depth_cache=None):
		""" constructor """
		super(ForwardCountDP, self).__init__(fst, trans_cache, state_depth_cache)
		self.alpha = None
		self._run()
		
	def _run(self):
		""" main algorithm """
		
		nstates = len(self.fst.states)
		## initialize the forward matrix
		# the sequence index starts with 1, 0 means "no part of the sequence seen so far"
		##self.alpha = numpy.zeros((nstates, 1))
		self.alpha = list()
		for i in range(0, self.max_seq_len() + 1):
			self.alpha.append(array.array("L", [0] * nstates))
		
		# we assume the starting state is always state 0
		self.alpha[0][0] = 1 

		for seqpos in range(1, self.max_seq_len() + 1):
			for stateid_to in self._state_depth_cache[seqpos]:
				value = 0
				for stateid_from in self._state_depth_cache[seqpos-1]:
					value += self._trans_cache[stateid_from][stateid_to] * self.alpha[seqpos-1][stateid_from]
				self.alpha[seqpos][stateid_to] = value
	
class BackwardCountDP(AbstractCountDP):
	def __init__(self, fst, trans_cache=None, state_depth_cache=None):
		""" constructor """
		super(BackwardCountDP, self).__init__(fst, trans_cache, state_depth_cache)
		self.beta = None
		self._run()

	
	def _run(self):
		""" main algorithm """
		
		## initialize the forward matrix
		# the sequence index starts with 1, 0 means "no part of the sequence seen so far"
		nstates = len(self.fst.states)
		self.beta = list()
		for stateid_from in range(0, self.max_seq_len() + 1):
			self.beta.append(array.array("L", [0] * nstates))
		
		# all final states are assigned 1
		for id in self._state_depth_cache[self.max_seq_len()]:
			self.beta[self.max_seq_len()][id] = 1 

		for seqpos in range(self.max_seq_len()-1, -1, -1):
			for stateid_from in self._state_depth_cache[seqpos]:
				value = 0
				for stateid_to in self._state_depth_cache[seqpos + 1]:
					value += self.beta[seqpos+1][stateid_to] * self._trans_cache[stateid_from][stateid_to]  
				self.beta[seqpos][stateid_from] = value

class PosteriorDecodingCountDP(AbstractCountDP):
	""" Dynamic programming algorithm for posterior decoding of counts """
	
	def __init__(self, fst):
		
		""" constructor """
		self.forward_algo = ForwardCountDP(fst)
		self.backward_algo = BackwardCountDP(fst,  
								   self.forward_algo.get_transition_cache(),
								   self.forward_algo.get_state_depth_cache())
		super(PosteriorDecodingCountDP, self).__init__(fst, 
												  self.forward_algo.get_transition_cache(),
												  self.forward_algo.get_state_depth_cache())        
		self.posterior = [dict() for i in range(0, self.max_seq_len())]
		self._run()
		
	def _run(self):
		""" main algorithm """
		nstates = len(self.fst.states)
		
		for pos in range(0, self.max_seq_len()):
			for state_from in self._state_depth_cache[pos]:
				all_transitions = self.fst.states[state_from].get_outgoing_transitions()
				for state_to in self._state_depth_cache[pos+1]:
					transitions = [t for t in all_transitions if t.state_to.id == state_to]
					fw = self.forward_algo.alpha[pos][state_from]
					bw = self.backward_algo.beta[pos+1][state_to]
					for trans in transitions:
						try:
							value = self.posterior[pos][trans.isymbol]
						except KeyError:
							value = 0
						value += fw * numpy.exp(-trans.weight) * bw
						self.posterior[pos][trans.isymbol] = value



class AbstractDP(object):
	def __init__(self, fst, trans_cache=None):
		""" constructor """
		self.fst = fst
		if trans_cache == None:
			self._build_cache()
		else:
			self._trans_cache = trans_cache
		assert(fst != None)

	def _get_transitions(self, state_from, state_to):
		""" returns all transitions between states with caching.
		Expects numerical state ids. """
		
		return self._trans_cache[state_from][state_to]
	
	def _build_cache(self):
		#self._trans_cache = numpy.empty( (len(self.fst.states), len(self.fst.states)), dtype="object")
		self._trans_cache = [None] * len(self.fst.states)
		self._trans_cache = [[i] * len(self.fst.states) for i in self._trans_cache]
		
		for i in range(0, len(self.fst.states)):
			from_transs = [t for t in self.fst.arcs if t.state_from.id == i]
			for j in range(0, len(self.fst.states)):
				transs = array.array('f', [t.weight for t in from_transs if t.state_to.id == j])
				self._trans_cache[i][j] = transs
				
				
		
class PosteriorDecodingDP(AbstractDP):
	""" Dynamic programming algorithm for posterior decoding """
	
	def __init__(self, fst):
		""" constructor """
		super(PosteriorDecodingDP, self).__init__(fst)        
		self.forward_algo = ForwardDP(self.fst)
		self.backward_algo = BackwardDP(self.fst, 
								   self.forward_algo.get_sequence_length(), 
								   self.forward_algo.get_transition_cache())
		self._trans_cache = self.forward_algo.get_transition_cache()
		self.posterior = [dict() for i in range(0, self.forward_algo.get_sequence_length())]
		self._run()
		
	def _run(self):
		""" main algorithm """
		
		
		for pos in range(0, self.forward_algo.get_sequence_length()):
			for state_from in range(0, len(self.fst.states)):
				for state_to in range(0, len(self.fst.states)):
					transitions = self._get_transitions(state_from, state_to)
					fw = self.forward_algo.alpha[state_from, pos]
					bw = self.backward_algo.beta[state_to, pos+1]
					for trans in transitions:
						try:
							value = self.posterior[pos][trans.isymbol]
						except KeyError:
							value = 0
						value += fw * numpy.exp(-trans.weight) * bw
						self.posterior[pos][trans.isymbol] = value
						
		
class ForwardDP(AbstractDP):
	""" Dynamic programming forward algorithm.
	This is not in log space and very generic. """
	
	def __init__(self, fst, trans_cache=None):
		""" constructor """
		super(ForwardDP, self).__init__(fst, trans_cache)
		self.alpha = None        
		self._run()
		
	def _run(self):
		""" main algorithm """
		
		nstates = len(self.fst.states)
		## initialize the forward matrix
		# the sequence index starts with 1, 0 means "no part of the sequence seen so far"
		##self.alpha = numpy.zeros((nstates, 1))
		self.alpha = list()
		self.alpha.append(array.array("f", [0] * nstates))
		
		# we assume the starting state is always state 0
		self.alpha[0][0] = 1 

		j=0
		while j<len(self.alpha):
			j+=1
			column = array.array("f", [0] * nstates)
			any_weight = False        
			for i in range(0, nstates):
				value = 0
				for k in range(0, nstates):
					transitions = self._trans_cache[k][i]
					for t in transitions:
						value += self.alpha[j-1][k] * math.exp(-t)  
				column[i] = value
				if value > 0.0: any_weight = True
			if any_weight:
				self.alpha.append(column)
	
	def get_sequence_length(self):
		return len(self.alpha)
	
	def get_transition_cache(self):
		return self._trans_cache
				
class BackwardDP(AbstractDP):
	""" Dynamic programming backward algorithm.
	This is not in log space and very generic. """
	
	def __init__(self, fst, seqlen, trans_cache=None):
		""" constructor """
		super(BackwardDP, self).__init__(fst, trans_cache)
		self.seqlen = seqlen
		self.beta = None
		self._run()
		
	def _run(self):
		""" main algorithm """
		
		## initialize the forward matrix
		# the sequence index starts with 1, 0 means "no part of the sequence seen so far"
		self.beta = numpy.zeros((len(self.fst.states), self.seqlen + 1))
		
		# all final states are assigned 1
		for state in self.fst.states:
			if state.is_final:
				self.beta[state.id, self.seqlen] = 1 

		for j in range(self.seqlen-1, -1, -1):
			for i in range(0, len(self.fst.states)):
				value = 0
				for k in range(0, len(self.fst.states)):
					transitions = self._get_transitions(i, k)
					for t in transitions:
						value += self.beta[k, j+1] * numpy.exp(-t)  
				self.beta[i,j] = value
		
		

class PathDepthFirstSearch(object):
	""" recursively walks through the FST and prints the paths.
	Make sure the FST does not contain loops, otherwise the function will not terminate. """
	
	def __init__(self, fst):
		'''
		Constructor
		'''
		self.fst = fst
	
	def get_paths(self):
		""" generator that returns paths """

		if self.fst.num_states()==0:
			return

		state = self.fst.start()
		next_arc = {}
		stack = []
		path = fstlib.core.Path()

		while state != None:
			arcs = [a for a in self.fst.arcs(state)]
			i = 0
			try:
				i = next_arc[state]
			except KeyError:
				pass
			
			if self.fst.num_arcs(state) > i: ## move forward
				arc = arcs[i]  ## get forward transition
				path.append(arc) ## save transition
				stack.append(state) ## put state on stack
				next_arc[state] = i + 1 ## increase trans counter by 1
				state = arc.nextstate ## move to next state
			else: ## move back
				if self.fst.is_final(state): ## end node
					path.finalWeight = self.fst.final(state)
					yield path.copy()

				next_arc[state] = 0    
				try:
					state = stack.pop()
					path.pop()
				except IndexError:
					state = None
				
				
							
class FeatureDepthFirstSearch(object):
	""" recursively walks through the FST and prints the paths.
	Make sure the FST does not contain loops, otherwise the function will not terminate. """
	
	def __init__(self, fst):
		'''
		Constructor
		'''
		self.fst = fst
		self.paths = []
		self.path_limit = None
		assert(fst != None)
	
	def run(self):
		""" starts the recursion """
		self.paths = []

		self._iteration()
	
	def _iteration(self):
		""" interative walk through the DAG """
		
		state = self.fst.states[0]
		suffix_cache = {}
		next_trans = {}
		stack = []
		counter = 0
		max_depth = 0
		while state != None:
			transitions_out = state.get_outgoing_transitions()
			i = 0
			try:
				i = next_trans[state]
			except KeyError:
				pass
			
			if len(transitions_out) > i: ## move forward
				max_depth += 1
				trans = transitions_out[i]  ## get forward transition
				stack.append(state) ## put state on stack
				next_trans[state] = i + 1 ## increase trans counter by 1
				state = trans.state_to ## move to next state                
				
			else: ## generate suffixes and jump back
				## do the jump back
				#print counter
				#counter += 1
				
				previous_state = None
				try:
					previous_state = stack.pop()
				except IndexError:
					pass
				
				if previous_state != None:
					prev_trans_id = next_trans[previous_state] - 1
					prev_trans = previous_state.get_outgoing_transitions()[prev_trans_id]
					suffixes_current = [[max_depth - 0, -1, -1]]
					try:
						suffixes_current = suffix_cache[state]
						#print suffixes_current
					except KeyError:
						pass
					#print "word: %s" % str(word)
					#print "concatenating..."
					#new_suffixes = [[word] + s for s in suffixes]
					#print "new suffixes:"
					#print [[str(t) for t in s] for s in new_suffixes]
						
					suffixes_prev = None
					try:
						suffixes_prev = suffix_cache[previous_state]
					except KeyError:
						suffixes_prev = []
						suffix_cache[previous_state] = suffixes_prev

					for suf in suffixes_current:
						new_suffix = [suf[0] - 1, suf[1], suf[2]]
						if prev_trans.osymbol == "X":
							assert( not(new_suffix[1] > -1 and new_suffix[2] > -1) )
							if new_suffix[2] == -1:
								new_suffix[2] = suf[0] - 2
							elif new_suffix[1] == -1:
								new_suffix[1] = suf[0]
								
						suffixes_prev.append(new_suffix)
					
					#print suffixes_prev
						
						#sc = sc + new_suffixes
						#print "sc:"
						#print [[str(t) for t in s] for s in sc]
					
						
				
				state = previous_state
		
		self.paths = suffix_cache[self.fst.states[0]]

def posterior_decoding(ifst, collapse_arcs=True):
	pd = PosteriorDecoding(ifst).run()
	conf = ConfusionNetworkFromLattice(pd).run(collapse_arcs)
	return conf

def map_log_to_real(x):
	return np.exp(-float(x))

def map_real_to_log(x):
	return -np.log(float(x))



import copy
import logging
import os

import fstlib
import numpy as np
import pandas as pd

from medicc import io, tools

# prepare logger 
logger = logging.getLogger(__name__)


def calculate_all_cn_events(tree, cur_df, alleles=['cn_a', 'cn_b'], normal_name='diploid',
                            wgd_x2=False, no_wgd=False, total_cn=False, max_wgd=1):
    """Create a DataFrame containing all copy-number events in the current data

    Args:
        tree (Bio.Phylo.Tree): Phylogenetic tree created by MEDICC2's tree reconstruction
        cur_df (pandas.DataFrame): DataFrame containing the copy-numbers of the samples and internal nodes
        alleles (list, optional): List of alleles. Defaults to ['cn_a', 'cn_b'].
        normal_name (str, optional): Name of the normal sample. Defaults to 'diploid'.

    Returns:
        pandas.DataFrame: Updated copy-number DataFrame
        pandas.DataFrame: DataFrame of copy-number events
    """
    
    cur_df[['is_gain', 'is_loss', 'is_wgd']] = False
    cur_df[alleles] = cur_df[alleles].astype(int)
    if tree == None:
        cur_df[['is_normal', 'is_clonal']] = False
        events = None
    else:

        events = pd.DataFrame(columns=['sample_id', 'chrom', 'start',
                                    'end', 'allele', 'type', 'cn_child'])

        clades = [x for x in tree.find_clades()]

        for clade in clades:
            if not len(clade.clades):
                continue
            if clade.name is None:
                clade = copy.deepcopy(clade)
                clade.name = normal_name
            for child in clade.clades:
                if child.branch_length == 0:
                    continue

                cur_df, cur_events = calculate_cn_events_per_branch(
                    cur_df, clade.name, child.name, alleles=alleles, wgd_x2=wgd_x2,
                    total_cn=total_cn, no_wgd=no_wgd, normal_name=normal_name, max_wgd=max_wgd)

                events = pd.concat([events, cur_events])

        events = events.reset_index(drop=True)

        is_normal = ~cur_df.unstack('sample_id')[['is_loss', 'is_gain', 'is_wgd']].any(axis=1)
        is_normal.name = 'is_normal'
        mrca = [x for x in tree.root.clades if x.name != normal_name][0].name
        is_clonal = ~cur_df.loc[cur_df.index.get_level_values('sample_id')!=mrca].unstack('sample_id')[['is_loss', 'is_gain', 'is_wgd']].any(axis=1)
        is_clonal.name = 'is_clonal'

        cur_df = cur_df.drop(['is_normal', 'is_clonal'], axis=1, errors='ignore')
        cur_df = (cur_df
                .join(is_normal, how='inner')
                .reorder_levels(['sample_id', 'chrom', 'start', 'end'])
                .sort_index()
                .join(is_clonal, how='inner')
                .reset_index())
        cur_df['chrom'] = tools.format_chromosomes(cur_df['chrom'])
        cur_df = (cur_df
                .set_index(['sample_id', 'chrom', 'start', 'end'])
                .sort_index())

        events = events.set_index(['sample_id', 'chrom', 'start', 'end'])

    return cur_df, events


def calculate_cn_events_per_branch(cur_df, parent_name, child_name, alleles=['cn_a', 'cn_b'],
                                   wgd_x2=False, total_cn=False, no_wgd=False, normal_name='diploid',
                                   max_wgd=1):
    """Calculate copy-number events for a single branch. Used in calculate_all_cn_events

    Args:
        cur_df (pandas.DataFrame): DataFrame containing the copy-numbers of the samples and internal nodes
        parent_name (str): Name of the parent sample
        child_name (str): Name of the child sample
        alleles (list, optional): List of alleles. Defaults to ['cn_a', 'cn_b'].

    Returns:
        pandas.DataFrame: Updated copy-number DataFrame
        pandas.DataFrame: DataFrame of copy-number events
    """

    cur_df = cur_df.copy()
    if len(np.setdiff1d(['is_gain', 'is_loss', 'is_wgd'], cur_df.columns)) > 0:
        cur_df[['is_gain', 'is_loss', 'is_wgd']] = False
    cur_df[alleles] = cur_df[alleles].astype(int)

    # TODO: load these outside of the function so they are not loaded every time
    asymm_fst, asymm_fst_nowgd, asymm_fst_1_wgd, asymm_fst_2_wgd, symbol_table = io.load_main_fsts(
        return_symbol_table=True)
    if total_cn:
        asymm_fst = io.read_fst(total_copy_numbers=True)
        asymm_fst_1_wgd = io.read_fst(total_copy_numbers=True, n_wgd=1)
        asymm_fst_2_wgd = None
    elif wgd_x2:
        asymm_fst = io.read_fst(wgd_x2=True)
        asymm_fst_1_wgd = io.read_fst(wgd_x2=True, n_wgd=1)
        asymm_fst_2_wgd = None
    elif no_wgd:
        asymm_fst = io.read_fst(no_wgd=True)
        asymm_fst_1_wgd = None
        asymm_fst_2_wgd = None


    events_df = pd.DataFrame(columns=['sample_id', 'chrom', 'start', 'end', 'allele', 'type', 'cn_child'])

    cur_parent_cn = cur_df.loc[parent_name, alleles].astype(int)
    cur_child_cn = cur_df.loc[child_name, alleles].astype(int)

    def get_int_chrom(x):
        if x == 'chrX':
            return 23
        elif x == 'chrY':
            return 24
        else:
            return int(x.split('chr')[-1])

    cur_chroms = cur_df.loc[normal_name].index.get_level_values(
        'chrom').map(get_int_chrom).values.astype(int)

    # 1. find total loss (loh)
    for allele in alleles:
        parent_loh = cur_parent_cn[allele].values == 0
        cur_loh = cur_child_cn.loc[~parent_loh, allele] == 0
        if cur_loh.sum() == 0:
            continue
        max_previous_cn = np.max(
            np.unique(cur_parent_cn.loc[~parent_loh, allele].loc[cur_loh]))
        for loh_height in np.arange(max_previous_cn):
            parent_loh = cur_parent_cn[allele].values == 0
            cur_loh = cur_child_cn.loc[~parent_loh, allele] == 0
            cn_changes = (cur_child_cn.loc[~parent_loh, allele] - 
                        cur_parent_cn.loc[~parent_loh, allele])

            # LOH detection
            cur_change_location = (cn_changes < 0)

            # + cur_chroms enables detection of chromosome boundaries
            cur_loss_labels_ = ((np.cumsum(np.concatenate([[0], np.diff(
                (cur_change_location.values + cur_chroms[~parent_loh]))])
                * cur_change_location.values) + 1)
                * cur_change_location.values)

            cur_loss_labels = np.zeros_like(cur_loss_labels_)
            for i, j in enumerate(np.setdiff1d(np.unique(cur_loss_labels_), [0])):
                cur_loss_labels[cur_loss_labels_ == j] = i + 1

            loh_regions = np.unique(cur_loss_labels[cur_loh])

            cur_events = (cur_child_cn
                            .loc[~parent_loh]
                            .reset_index()
                            .loc[np.array([np.argmax(cur_loss_labels == val) for val in np.setdiff1d(np.unique(loh_regions), [0])])]
                            [['chrom', 'start', 'end', allele]].values)

            # adjust ends
            cur_events[:, 2] = (cur_child_cn
                                .loc[~parent_loh]
                                .reset_index()
                                .loc[np.array([len(cur_loss_labels) - np.argmax(cur_loss_labels[::-1] == val) - 1 for val in np.setdiff1d(np.unique(loh_regions), [0])])]
                                ['end'].values)
            
            cur_ind = np.arange(len(events_df), len(events_df)+len(cur_events))
            events_df = pd.concat([events_df, pd.DataFrame(index=cur_ind)])
            events_df.loc[cur_ind, 'sample_id'] = child_name
            events_df.loc[cur_ind, 'allele'] = allele
            events_df.loc[cur_ind, 'type'] = 'loss'
            events_df.loc[cur_ind, 'cn_child'] = max_previous_cn - loh_height - 1
            events_df.loc[cur_ind, ['chrom', 'start', 'end']] = cur_events[:, :3]

            cur_parent_cn.loc[cn_changes.loc[np.stack([cur_loss_labels == val for val in loh_regions]).any(axis=0)].index, allele] -= 1

            cur_df.loc[cur_df.loc[[child_name]].loc[~parent_loh].loc[cur_change_location.values].index, 'is_loss'] = True

    loh_pos = (cur_parent_cn == 0)

    # 2. WGDs
    # only check if >30% of is gained
    wgd_candidate_threshold = 0.3

    widths = cur_df.loc[[child_name]].eval('end+1-start')
    fraction_gain = ((cur_df.loc[child_name, alleles] > 1).astype(int).sum(axis=1) * widths.loc[child_name]
                        ).sum() / (2 * widths.loc[child_name].sum())
    parent_fsa = fstlib.factory.from_string('X'.join(['X'.join(["".join(x.astype('str')) for _, x in cur_df.loc[parent_name, alleles][allele].groupby('chrom')]) for allele in alleles]),
                                            arc_type="standard",
                                            isymbols=symbol_table,
                                            osymbols=symbol_table)
    child_fsa = fstlib.factory.from_string('X'.join(['X'.join(["".join(x.astype('str')) for _, x in cur_df.loc[child_name, alleles][allele].groupby('chrom')]) for allele in alleles]),
                                            arc_type="standard",
                                            isymbols=symbol_table,
                                            osymbols=symbol_table)

    score_wgd = float(fstlib.score(asymm_fst, parent_fsa, child_fsa))
    fraction_double_gain = (((cur_df.loc[child_name, alleles] > 2)
                            .astype(int)
                            .sum(axis=1)
                            * widths.loc[child_name]
                            ).sum() / widths.loc[child_name].sum())
    if not no_wgd and fraction_gain > wgd_candidate_threshold:
        if wgd_x2:
            # double wgd
            if (fraction_double_gain > wgd_candidate_threshold) and (float(fstlib.score(asymm_fst_1_wgd, parent_fsa, child_fsa)) != score_wgd):
                cur_parent_cn = 4 * cur_parent_cn
                events_df.loc[len(events_df.index)] = [child_name, 'chr0', cur_df.index.get_level_values('start').min(),
                                                cur_df.index.get_level_values('end').max(), 'both', 'wgd', 0]
                events_df.loc[len(events_df.index)] = [child_name, 'chr0', cur_df.index.get_level_values('start').min(),
                                                    cur_df.index.get_level_values('end').max(), 'both', 'wgd', 0]
                cur_df.loc[child_name, 'is_wgd'] = True
            # single wgd
            elif float(fstlib.score(asymm_fst_nowgd, parent_fsa, child_fsa)) != score_wgd:
                cur_parent_cn = 2 * cur_parent_cn
                events_df.loc[len(events_df.index)] = [child_name, 'chr0', cur_df.index.get_level_values('start').min(),
                                                cur_df.index.get_level_values('end').max(), 'both', 'wgd', 0]
                cur_df.loc[child_name, 'is_wgd'] = True

        elif total_cn:
            # single wgd
            if float(fstlib.score(asymm_fst_nowgd, parent_fsa, child_fsa)) != score_wgd:
                cur_parent_cn[~loh_pos] = cur_parent_cn[~loh_pos] + 2
                events_df.loc[len(events_df.index)] = [child_name, 'chr0', cur_df.index.get_level_values('start').min(),
                                                cur_df.index.get_level_values('end').max(), 'both', 'wgd', 0]
                cur_df.loc[child_name, 'is_wgd'] = True

        else:
            # triple wgd
            if max_wgd >= 3 and (fraction_double_gain > wgd_candidate_threshold) and (float(fstlib.score(asymm_fst_2_wgd, parent_fsa, child_fsa)) != score_wgd):
                cur_parent_cn[~loh_pos] = cur_parent_cn[~loh_pos] + 3
                events_df.loc[len(events_df.index)] = [child_name, 'chr0', cur_df.index.get_level_values('start').min(),
                                                cur_df.index.get_level_values('end').max(), 'both', 'wgd', 0]
                events_df.loc[len(events_df.index)] = [child_name, 'chr0', cur_df.index.get_level_values('start').min(),
                                                    cur_df.index.get_level_values('end').max(), 'both', 'wgd', 0]
                events_df.loc[len(events_df.index)] = [child_name, 'chr0', cur_df.index.get_level_values('start').min(),
                                                    cur_df.index.get_level_values('end').max(), 'both', 'wgd', 0]
                cur_df.loc[child_name, 'is_wgd'] = True
            # double wgd
            elif max_wgd >= 2 and (fraction_double_gain > wgd_candidate_threshold) and (float(fstlib.score(asymm_fst_1_wgd, parent_fsa, child_fsa)) != score_wgd):
                cur_parent_cn[~loh_pos] = cur_parent_cn[~loh_pos] + 2
                events_df.loc[len(events_df.index)] = [child_name, 'chr0', cur_df.index.get_level_values('start').min(),
                                                cur_df.index.get_level_values('end').max(), 'both', 'wgd', 0]
                events_df.loc[len(events_df.index)] = [child_name, 'chr0', cur_df.index.get_level_values('start').min(),
                                                    cur_df.index.get_level_values('end').max(), 'both', 'wgd', 0]
                cur_df.loc[child_name, 'is_wgd'] = True
            # single wgd
            elif max_wgd >= 1 and float(fstlib.score(asymm_fst_nowgd, parent_fsa, child_fsa)) != score_wgd:
                cur_parent_cn[~loh_pos] = cur_parent_cn[~loh_pos] + 1
                events_df.loc[len(events_df.index)] = [child_name, 'chr0', cur_df.index.get_level_values('start').min(),
                                                cur_df.index.get_level_values('end').max(), 'both', 'wgd', 0]
                cur_df.loc[child_name, 'is_wgd'] = True

    # 3. losses and gains
    for allele in alleles:

        cn_changes = (cur_child_cn[allele] - cur_parent_cn[allele]).values
        all_cn_change_vals = np.unique(cn_changes)

        cur_df.loc[child_name, 'is_loss'] = np.logical_or(cur_df.loc[child_name, 'is_loss'].values,
                                                          cn_changes < 0)
        cur_df.loc[child_name, 'is_gain'] = np.logical_or(cur_df.loc[child_name, 'is_gain'].values,
                                                          cn_changes > 0)

        # enumerate over all possible change values
        all_cn_change_vals = np.setdiff1d(np.arange(np.min([np.min(all_cn_change_vals), 0]), np.max(all_cn_change_vals)+1), [0])
        for cur_cn_change in all_cn_change_vals[np.argsort(np.abs(all_cn_change_vals))[::-1]]:
            cur_event = 'gain' if cur_cn_change > 0 else 'loss'

            cur_change_location = ((cur_child_cn.loc[~loh_pos[allele], allele] -
                        cur_parent_cn.loc[~loh_pos[allele], allele]) == cur_cn_change)

            # + cur_chroms enables detection of chromosome boundaries
            event_labels_ = ((np.cumsum(np.concatenate([[0], np.diff(
                (cur_change_location.values + cur_chroms[~loh_pos[allele]]))])
                * cur_change_location.values) + 1)
                * cur_change_location.values)

            # Label events starting at 1
            event_labels = np.zeros_like(event_labels_)
            for i, j in enumerate(np.setdiff1d(np.unique(event_labels_), [0])):
                event_labels[event_labels_ == j] = i + 1

            cur_events = (cur_child_cn
                          .loc[~loh_pos[allele]]
                          .reset_index()
                          .loc[np.array([np.argmax(event_labels == val) for val in np.setdiff1d(np.unique(event_labels), [0])])]
                          [['chrom', 'start', 'end', allele]].values)

            # adjust ends
            cur_events[:, 2] = (cur_child_cn
                                .loc[~loh_pos[allele]]
                                .reset_index()
                                .loc[np.array([len(event_labels) - np.argmax(event_labels[::-1] == val) - 1 for val in np.setdiff1d(np.unique(event_labels), [0])])]
                                ['end'].values)

            cur_ind = np.arange(len(events_df), len(events_df)+len(cur_events))
            events_df = pd.concat([events_df, pd.DataFrame(index=cur_ind)])
            events_df.loc[cur_ind, 'sample_id'] = child_name
            events_df.loc[cur_ind, 'allele'] = allele
            events_df.loc[cur_ind, 'type'] = cur_event
            events_df.loc[cur_ind, 'cn_child'] = cur_events[:, 3]
            events_df.loc[cur_ind, ['chrom', 'start', 'end']] = cur_events[:, :3]

            cur_child_cn.loc[np.intersect1d(loh_pos.loc[~loh_pos[allele]].index,
                                            cur_change_location.loc[cur_change_location].index), allele] += (1 if (cur_cn_change < 0) else -1)

    events_df['chrom'] = tools.format_chromosomes(events_df['chrom'])
    events_df = (events_df[['sample_id', 'allele', 'chrom', 'start', 'end', 'type', 'cn_child']]
                 .reset_index(drop=True)
                 .sort_values(['sample_id', 'allele', 'chrom', 'start', 'end', 'type', 'cn_child']))

    return cur_df, events_df


def overlap_events(events_df=None, output_df=None, tree=None, overlap_threshold=0.9,
                   chromosome_bed='default', regions_bed='default',
                   replace_loh_with_loss=True, alleles=['cn_a', 'cn_b'],
                   replace_both_arms_with_chrom=True, normal_name='diploid'):
    """Overlap copy-number events with regions of interest

    Args:
        events_df (pandas.DataFrame, optional): All copy-number events. Defaults to None.
        output_df (pandas.DataFrame, optional): DataFrame containing all copy-numbers. Defaults to None.
        tree (Bio.Phylo.Tree, optional): Phylogenetic tree. Defaults to None.
        overlap_threshold (float, optional): Threshold above which an overlap is considered. Defaults to 0.9.
        chromosome_bed (str, optional): Name of BED file containing chromosome arm information. Defaults to 'none'.
        regions_bed (str, optional): Name of BED file containing regions of interest. Defaults to 'none'.
        replace_loh_with_loss (bool, optional): If True, loh is considered like a normal loss. Defaults to True.
        alleles (list, optional): List of alleles. Defaults to ['cn_a', 'cn_b'].
        replace_both_arms_with_chrom (bool, optional): If True, an event in the p- and q-arm of a chromosome will be displayed as a single event. Defaults to True.
        normal_name (str, optional): Name of normal sample. Defaults to 'diploid'.

    Returns:
        pandas.DataFrame: DataFrame with events concerning the regions of interest
    """
    try:
        import pyranges as pr
    except ImportError:
        raise ImportError("You have to install pyranges to overlap events with regions of interest") from None       

    if chromosome_bed == 'none':
        chromosome_bed = None
    elif chromosome_bed == 'default':
        chromosome_bed = os.path.join(os.path.dirname(os.path.realpath(__file__)),
                                      "objects", "hg38_chromosome_arms.bed")
    if regions_bed == 'none':
        regions_bed = None
    elif regions_bed == 'default':
        regions_bed = os.path.join(os.path.dirname(os.path.realpath(__file__)),
                                   "objects", "Davoli_2013_TSG_OG_genes.bed")

    all_events = pd.DataFrame(columns=['Chromosome', 'Start', 'End', 'name', 'NumberOverlaps',
                                       'FractionOverlaps', 'event', 'branch']).set_index(['Chromosome', 'Start', 'End'])

    if events_df is None:
        if output_df is None or tree is None:
            raise MEDICCEventReconstructionError("Either events_df or df and tree has to be specified")
        _, events_df = calculate_all_cn_events(tree, output_df, alleles=alleles, normal_name=normal_name)
    if replace_loh_with_loss:
        events_df.loc[events_df['type'] == 'loh', 'type'] = 'loss'

    # Read chromosome regions and other regions
    if chromosome_bed is None and regions_bed is None:
        raise MEDICCEventReconstructionError("Either chromosome_bed or regions_bed has to be specified")

    chr_arm_regions = None
    if chromosome_bed is not None:
        logger.debug(f'Overlap with chromosomes bed file {chromosome_bed}')
        chr_arm_regions = io.read_bed_file(chromosome_bed)
        whole_chromosome = chr_arm_regions.groupby('Chromosome').min()
        whole_chromosome['End'] = chr_arm_regions.groupby('Chromosome')['End'].max()
        whole_chromosome['name'] = whole_chromosome.index
        chr_arm_regions = pd.concat([chr_arm_regions, whole_chromosome.reset_index()]).sort_values('Chromosome')
        chr_arm_regions = pr.PyRanges(chr_arm_regions)

    regions = None
    if regions_bed is not None:
        logger.debug(f'Overlap with regions bed file {regions_bed}')
        regions = []
        if isinstance(regions_bed, list) or isinstance(regions_bed, tuple):
            for f in regions_bed:
                regions.append(pr.PyRanges(io.read_bed_file(f)))
        else:
            regions.append(pr.PyRanges(io.read_bed_file(regions_bed)))

    # Add WGD
    for ind, _ in events_df.loc[events_df['type'] == 'wgd'].iterrows():
        all_events.loc[('all', '0', '0')] = ['WGD', 1, 1., 'WGD', ind[0]]

    for cur_branch in events_df.index.get_level_values('sample_id').unique():
        for allele in alleles:
            cur_events_df = events_df.loc[cur_branch]
            cur_events_df = cur_events_df.loc[cur_events_df['allele']==allele]
            for event_type in ['gain', 'loss'] if replace_loh_with_loss else ['gain', 'loh', 'loss']:
                cur_events_ranges = pr.PyRanges(cur_events_df.loc[cur_events_df['type'] == event_type].reset_index(
                ).rename({'chrom': 'Chromosome', 'start': 'Start', 'end': 'End'}, axis=1))

                # Calculate chromosomal events
                if chr_arm_regions is not None:
                    chr_events = overlap_regions(
                        chr_arm_regions, cur_events_ranges, event_type, cur_branch, overlap_threshold)
                    # remove arms if the whole chromosome is in there
                    if replace_both_arms_with_chrom and len(chr_events) > 0:
                        chr_events = chr_events[~chr_events['name'].isin(np.concatenate(
                            [[name + 'p', name + 'q'] if ('q' not in name and 'p' not in name) else [] for name in chr_events['name']]))]
                    all_events = pd.concat([all_events, chr_events])

                # Calculate other events
                if regions is not None:
                    for region in regions:
                        chr_events = overlap_regions(
                            region, cur_events_ranges, event_type, cur_branch, overlap_threshold)
                        all_events = pd.concat([all_events, chr_events])

    all_events['final_name'] = all_events['name'].apply(lambda x: x.split(
        'chr')[-1]) + all_events['event'].apply(lambda x: ' +' if x == 'gain' else (' -' if x == 'loss' else (' 0' if x == 'loh' else '')))

    all_events = all_events.set_index(['branch', 'name']).drop('NumberOverlaps', axis=1)
    all_events = all_events.reset_index().set_index('branch')
    
    return all_events


def overlap_regions(region, cur_events_ranges, event, branch, overlap_threshold):

    # Workaround for bug in PyRanges for numpy version 1.20 and higher
    np.long = np.int_

    cur_events_overlaps = region.coverage(cur_events_ranges).as_df()
    cur_events_overlaps = cur_events_overlaps.loc[cur_events_overlaps['FractionOverlaps']
                                                > overlap_threshold]
    cur_events_overlaps = cur_events_overlaps.set_index(['Chromosome', 'Start', 'End'])
    cur_events_overlaps['event'] = event
    cur_events_overlaps['branch'] = branch

    return cur_events_overlaps


class MEDICCEventReconstructionError(Exception):
    pass



import logging

import fstlib
import numpy as np

from medicc import tools

SYMBOL_GAP = 'gap'

# prepare logger
logger = logging.getLogger(__name__)


def create_symbol_table(max_cn=8, separator='X'):
    alphabet = [tools.int2hex(x) for x in range(max_cn+1)]
    symbol_table = fstlib.SymbolTable()
    symbol_table.add_symbol(SYMBOL_GAP)
    for s in alphabet:
        symbol_table.add_symbol(s)

    if separator is not None and separator != '':
        symbol_table.add_symbol(separator)
    return symbol_table


def _get_int_cns_from_symbol_table(symbol_table, separator='X'):
    cns = {x[1]: tools.hex2int(x[1]) for x in symbol_table if x[0] != 0 and x[1] != separator}
    return cns


def create_1step_del_fst(symbol_table, separator='X', exclude_zero=False, w_stay=0, w_open=1, w_extend=0):

    cns = _get_int_cns_from_symbol_table(symbol_table, separator)

    myfst = fstlib.Fst(arc_type='standard')
    myfst.set_input_symbols(symbol_table)
    myfst.set_output_symbols(symbol_table)

    myfst.add_states(2)
    myfst.set_start(0)
    myfst.set_final(0, 0)
    myfst.set_final(1, 0)

    myfst.add_arcs(0, [(s, s, w_stay, 0) for s in cns.keys()])
    if separator is not None and separator != '':
        myfst.add_arc(0, (separator, separator, 0, 0))  # add separator
    if exclude_zero:
        myfst.add_arcs(0, [(s, t, w_open, 1) for s in cns.keys()
                           for t in cns.keys() if (cns[s]-cns[t]) == 1 and t != '0'])  # 0->1
        myfst.add_arcs(1, [(s, t, w_extend, 1) for s in cns.keys() for t in cns.keys()
                           if (cns[s]-cns[t]) == 1 and t != '0'])  # extend an open window
    else:
        myfst.add_arcs(0, [(s, t, w_open, 1) for s in cns.keys()
                           for t in cns.keys() if (cns[s]-cns[t]) == 1])  # 0->1
        myfst.add_arcs(1, [(s, t, w_extend, 1) for s in cns.keys()
                           for t in cns.keys() if (cns[s]-cns[t]) == 1])  # extend an open window
    if '0' in cns.keys():
        myfst.add_arc(1, ('0', '0', 0, 1))
    myfst.add_arcs(1, [(s, s, w_stay, 0) for s in cns.keys() if s != '0'])  # return
    if separator is not None and separator != '':
        myfst.add_arc(1, (separator, separator, 0, 0))

    return myfst


def create_loh_fst(symbol_table, separator='X'):

    cns = _get_int_cns_from_symbol_table(symbol_table, separator)

    myfst = fstlib.Fst(arc_type='standard')
    myfst.set_input_symbols(symbol_table)
    myfst.set_output_symbols(symbol_table)

    myfst.add_states(9)
    myfst.set_start(0)
    for i in range(9):
        myfst.set_final(i, 0)

    ## match 0 -> 0
    myfst.add_arcs(0, [(s, s, 0, 0) for s in cns.keys()])
    if separator is not None and separator != '':
        myfst.add_arc(0, (separator, separator, 0, 0))  # add separator

    ## others
    for state in range(1, 9):
        cost = state
        myfst.add_arcs(0, [(s, t, cost, state) for s in cns.keys()
                           for t in cns.keys() if (cns[s]-cns[t]) == cost and t == '0'])
        myfst.add_arcs(state, [(s, t, 0, state) for s in cns.keys()
                               for t in cns.keys() if (cns[s]-cns[t]) <= cost and t == '0'])
        myfst.add_arcs(state, [(s, t, cns[s]-cns[t]-cost, cns[s]-cns[t]) for s in cns.keys()
                               for t in cns.keys() if (cns[s]-cns[t]) > cost and t == '0'])
        myfst.add_arcs(state, [(s, s, 0, 0) for s in cns.keys()])
        if separator is not None and separator != '':
            myfst.add_arc(state, (separator, separator, 0, 0))

    myfst = fstlib.encode_determinize_minimize(myfst)
    
    return myfst


def create_1step_WGD_fst(symbol_table, separator='X', wgd_cost=1, minimize=True, wgd_x2=False,
                         total_cn=False):

    if total_cn:
        wgd_distance = 2.
    else:
        wgd_distance = 1.

    cns = _get_int_cns_from_symbol_table(symbol_table, separator)

    W = fstlib.Fst()
    W.set_input_symbols(symbol_table)
    W.set_output_symbols(symbol_table)
    W.add_states(3)
    W.set_start(0)
    W.set_final(0, 0)
    W.set_final(1, 0)
    W.set_final(2, 0)
    if wgd_x2:
        W.add_arcs(0, [(s, t, wgd_cost, 1) for s in cns.keys()
                    for t in cns.keys() if (s != '0') and ((cns[t]/cns[s]) == 2.)])
    else:
        W.add_arcs(0, [(s, t, wgd_cost, 1) for s in cns.keys()
                    for t in cns.keys() if (s != '0') and ((cns[t]-cns[s]) == wgd_distance)])

    W.add_arc(1, ('0', '0', 0, 1))
    if wgd_x2:
        W.add_arcs(1, [(s, t, 0, 1) for s in cns.keys()
                    for t in cns.keys() if (s != '0') and ((cns[t]/cns[s]) == 2.)])
    else:
        W.add_arcs(1, [(s, t, 0, 1) for s in cns.keys()
                    for t in cns.keys() if (s != '0') and ((cns[t]-cns[s]) == wgd_distance)])
    W.add_arc(0, ('0', '0', 0, 0))
    if separator is not None and separator != '':
        W.add_arc(0, (separator, separator, 0, 0))
        W.add_arc(1, (separator, separator, 0, 1))
    W.add_arc(1, ('0', '0', 0, 1))

    W.add_arcs(0, [(s, s, 0, 2) for s in cns.keys() if s != '0'])
    W.add_arcs(2, [(s, s, 0, 2) for s in cns.keys()])
    W.add_arc(2, ('0', '0', 0, 2))
    if separator is not None and separator != '':
        W.add_arc(2, (separator, separator, 0, 2))
    W.arcsort('olabel')
    if minimize:
        W = fstlib.encode_determinize_minimize(W)

    return W


def create_nstep_fst(n, one_step_fst, minimize=True):
    # Extend 1step FST
    nstep_fst = one_step_fst
    nstep_fst.arcsort(sort_type='olabel')

    for _ in range(n):
        nstep_fst = fstlib.compose(nstep_fst, one_step_fst)
        if minimize:
            nstep_fst = fstlib.encode_determinize_minimize(nstep_fst)
        nstep_fst.arcsort(sort_type='olabel')

    return nstep_fst


def create_copynumber_fst(symbol_table, sep='X', enable_wgd=False, wgd_cost=1, 
                          max_num_wgds=3, wgd_x2=False, output_all=False, total_cn=False,
                          exact=True, max_pre_wgd_losses=8, exact_wgd=False):
    """ Creates the tree FST T which computes the asymmetric MED.
    The current creation is based on a trade-off. In the absence of WGDs, the FST is exact wr.t.
    combined LOH-losses (i.e 21 -> 10 is counted as one event), however, in the presence of WGDs, 
    for performance reasons losses and LOHs are counted separately.

    For `exact_wgd=True`, the FST is exact wr.t. combined LOH-losses even in the presence of WGDs 
    but due to the exponential growth of the FST, it is not feasible in most cases. (Example where 
    this is required: 22X1X1X1 -> 10X2X2X2, 4 events with exact_wgd and 5 otherwise.)

    For `exact=False` the legacy version is created which never combines losses and LOHs.
    """
    n = len(_get_int_cns_from_symbol_table(symbol_table, sep))

    L_1step = create_1step_del_fst(symbol_table, sep, exclude_zero=True)
    L = create_nstep_fst(n-1, L_1step)
    LG = fstlib.encode_determinize_minimize(L*~L)
    G = ~L
    L_LOH_1step = create_1step_del_fst(symbol_table, sep, exclude_zero=False)
    L_LOH = create_nstep_fst(max_pre_wgd_losses-1, L_LOH_1step)
    LOH = create_loh_fst(symbol_table, sep)
    
    if enable_wgd:
        W1step = create_1step_WGD_fst(symbol_table, sep, wgd_cost=wgd_cost,
                                      minimize=False, wgd_x2=wgd_x2, total_cn=total_cn)
        n = int(min(max_num_wgds, np.floor(np.log2(n))))
        if n > 1:
            W = create_nstep_fst(n-1, W1step)
        else:
            W = W1step
        if exact_wgd:
            T = L_LOH * W * LG
        elif exact:
            T = ((LOH * W * LG) + fstlib.encode_determinize_minimize(L_LOH * G)).rmepsilon()
        else:
            # legacy version
            T = LOH * W * LG
    else:
        if exact:
            T = fstlib.encode_determinize_minimize(L_LOH*G)
        else:
            # legacy version
            T = LOH * LG
        W = None

    if output_all:
        return {'T': T, 'LOH': LOH, 'W': W, 'L': L, 'L_LOH': L_LOH, 'G': G, 'LG': LG}
    else:
        return T



#cython: language_level=3

from libcpp.vector cimport vector

cimport cpywrapfst as fst

cdef extern from "cops.h" nogil:
	cdef void shortest_path(const fst.FstClass &, const fst.FstClass &, const fst.FstClass &, fst.MutableFstClass *)
	cdef fst.WeightClass shortest_distance_std(const fst.FstClass &, const fst.FstClass &, const fst.FstClass &)
	cdef fst.WeightClass shortest_distance_log(const fst.FstClass &, const fst.FstClass &, const fst.FstClass &)
	cdef fst.WeightClass kernel_score_std_impl(const fst.FstClass &, const fst.FstClass &, const fst.FstClass &)


import logging
import re

import Bio
import fstlib
import numpy as np
import pandas as pd

logger = logging.getLogger('medicc.io')

def set_sequences_on_tree_from_df(tree: Bio.Phylo.BaseTree, df: pd.DataFrame, clear_before=True):
    """DEPRECATED!
    
    Set sequences on tree from dataframe

    Args:
        tree (Bio.Phylo.BaseTree): Tree to set sequences on
        df (pd.DataFrame): DataFrame with copy number information
        clear_before (bool, optional): Clear old sequences. Defaults to True.
    """
    raise DeprecationWarning("This function is deprecated and is not used anymore. ")

    if not hasattr(tree.root, 'sequences'):
        tree = tree.as_phyloxml()
    for clade in tree.find_clades():
        if clear_before:
            clade.sequences.clear()
        for label, data in df.iteritems():
            try:
                clade.sequences.append(
                    Bio.Phylo.PhyloXML.Sequence(
                        name='X'.join(data.loc[clade.name].groupby('chrom').apply(lambda x: ''.join(x))),
                        symbol=label.upper())) 
            except KeyError:
                pass


def fsa_to_string(fsa):
    fsa_string = fstlib.tools.strings(fsa).string[0]
    return fsa_string

def hex2int(x):
    return int(x, 16)

def int2hex(x):
    return hex(x)[2:]

def format_chromosomes(ds):
    """ Expects pandas Series with chromosome names. 
    The goal is to take recognisalbe chromosome names, i.e. chr4 or chrom3 and turn them into chr3 format.
    If the chromosomes names are not recognized, return them unchanged."""

    ds = ds.astype('str')
    pattern = re.compile(r"(chr|chrom)?(_)?(0)?((\d+)|X|Y)", flags=re.IGNORECASE)
    matches = ds.apply(pattern.match)
    matchable = ~matches.isnull().any()
    if matchable:
        newchr = matches.apply(lambda x:"chr%s" % x[4].upper())
        numchr = matches.apply(lambda x:int(x[5]) if x[5] is not None else -1)
        chrlevels = np.sort(numchr.unique())
        chrlevels = np.setdiff1d(chrlevels, [-1])
        chrcats = ["chr%d" % i for i in chrlevels]
        if 'chrX' in list(newchr):
            chrcats += ['chrX',]
        if 'chrY' in list(newchr):
            chrcats += ['chrY',]
        newchr = pd.Categorical(newchr, categories=chrcats)
    else:
        logger.warn("Could not match the chromosome labels. Rename the chromosomes according chr1, "
                    "chr2, ... to avoid potential errors."
                    "Current format: {}".format(ds.unique()))
        newchr = pd.Categorical(ds, categories=ds.unique())
    assert not newchr.isna().any(), "Could not reformat chromosome labels. Rename according to chr1, chr2, ..."
    return newchr


def next_prime(N):
    '''calculate a prime number p with p >= N'''
    def is_prime(x):
        return all(x % i for i in range(3, x))

    cur_p = int(np.ceil(np.sqrt(N)))
    if cur_p % 2 == 0:
        cur_p += 1
    while not is_prime(cur_p):
        cur_p += 2
    return cur_p


def create_parallelization_groups(number_samples):
    '''Create subgroups of the samples to perform parallel runs of the pairwise distances matrix 
    calculations. 

    Args:
        number_samples (int): Total number of samples

    Returns:
        list of lists: Lists, each one containing up to p indices


    Method implemented as proposed in:
    Emmanuel Sapin, Matthew C Keller, Novel approach for parallelizing pairwise comparison problems as applied to detecting segments identical by decent in whole-genome data, 
    Bioinformatics, 2021;, btab084, https://doi.org/10.1093/bioinformatics/btab084
    '''
    p = next_prime(number_samples)
    p_matrix = np.arange(p**2).reshape((p, p)).T
    p_matrix[p_matrix >= number_samples] = -1

    p_matrix_permutations = [p_matrix]
    for _ in range(p-1):
        p_matrix_permutations.append(np.array([np.roll(p_matrix_permutations[-1][:, i], -i) for i in range(p)]).T)

    groups = [*p_matrix.T] + [x for cur_p in p_matrix_permutations for x in [*cur_p]]
    # remove -1 entries (numbers larger than number_samples)
    groups = [group[group >= 0] for group in groups]
    groups = [group for group in groups if len(group) > 1]

    return groups


def total_pdm_from_parallel_pdms(sample_labels, parallel_pdms):
    total_pdm = pd.DataFrame(index=sample_labels, columns=sample_labels, dtype=float)

    for cur_pdm in parallel_pdms:
        total_pdm.loc[cur_pdm.index, cur_pdm.index] = cur_pdm

    if total_pdm.isna().sum().sum() != 0:
        raise ValueError('Something went wrong with these indices:\n{}'.format(
            np.where(total_pdm.isna())))

    return total_pdm


def create_diploid_fsa(fst, total_copy_numbers=False):

    # Create diploid FSA
    diploid_fsa = fstlib.Fst()
    diploid_fsa.set_input_symbols(fst.input_symbols())
    diploid_fsa.set_output_symbols(fst.output_symbols())
    diploid_fsa.add_state()
    diploid_fsa.set_start(0)
    diploid_fsa.set_final(0, 0)
    if total_copy_numbers:
        diploid_fsa.add_arc(0, ('2', '2', 0, 0))
    else:
        diploid_fsa.add_arc(0, ('1', '1', 0, 0))
    diploid_fsa.add_arc(0, ('X', 'X', 0, 0))

    return diploid_fsa
